# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from collections import defaultdict
from functools import cached_property
from pathlib import Path
from typing import Optional

from anemoi.utils.checkpoints import load_metadata
from earthkit.data.utils.dates import to_datetime

from .metadata import Metadata

LOG = logging.getLogger(__name__)


def _download_huggingfacehub(huggingface_config) -> str:
    """Download model from huggingface"""
    try:
        from huggingface_hub import hf_hub_download
        from huggingface_hub import snapshot_download
    except ImportError as e:
        raise ImportError("Could not import `huggingface_hub`, please run `pip install huggingface_hub`.") from e

    if isinstance(huggingface_config, str):
        huggingface_config = {"repo_id": huggingface_config}

    if "filename" in huggingface_config:
        return str(hf_hub_download(**huggingface_config))

    repo_path = Path(snapshot_download(**huggingface_config))
    ckpt_files = list(repo_path.glob("*.ckpt"))

    if len(ckpt_files) == 1:
        return str(ckpt_files[0])
    else:
        raise ValueError(
            f"None or Multiple ckpt files found in repo, {ckpt_files}.\nCannot pick one to load, please specify `filename`."
        )


class Checkpoint:
    """Represents an inference checkpoint."""

    def __init__(self, path, *, patch_metadata=None):
        self._path = path
        self.patch_metadata = patch_metadata

    def __repr__(self):
        return f"Checkpoint({self.path})"

    @cached_property
    def path(self) -> str:
        import json

        try:
            path = json.loads(self._path)
        except Exception:
            path = self._path

        if isinstance(path, (Path, str)):
            return str(path)
        elif isinstance(path, dict):
            if "huggingface" in path:
                return _download_huggingfacehub(path["huggingface"])
            pass
        raise TypeError(f"Cannot parse model path: {path}. It must be a path or dict")

    @cached_property
    def _metadata(self):
        try:
            result = Metadata(*load_metadata(self.path, supporting_arrays=True))
        except Exception as e:
            LOG.warning("Version for not support `supporting_arrays` (%s)", e)
            result = Metadata(load_metadata(self.path))

        if self.patch_metadata:
            LOG.warning("Patching metadata with %r", self.patch_metadata)
            result.patch(self.patch_metadata)

        return result

    ###########################################################################
    # Forwards used by the runner
    # We do not want to expose the metadata object directly
    # We do not use `getattr` to avoid exposing all methods and make debugging
    # easier
    ###########################################################################

    @property
    def timestep(self):
        return self._metadata.timestep

    @property
    def precision(self):
        return self._metadata.precision

    @property
    def number_of_grid_points(self):
        return self._metadata.number_of_grid_points

    @property
    def number_of_input_features(self):
        return self._metadata.number_of_input_features

    @property
    def variable_to_input_tensor_index(self):
        return self._metadata.variable_to_input_tensor_index

    @property
    def model_computed_variables(self):
        return self._metadata.model_computed_variables

    @property
    def typed_variables(self):
        return self._metadata.typed_variables

    @property
    def diagnostic_variables(self):
        return self._metadata.diagnostic_variables

    @property
    def prognostic_variables(self):
        return self._metadata.prognostic_variables

    @property
    def prognostic_output_mask(self):
        return self._metadata.prognostic_output_mask

    @property
    def prognostic_input_mask(self):
        return self._metadata.prognostic_input_mask

    @property
    def output_tensor_index_to_variable(self):
        return self._metadata.output_tensor_index_to_variable

    @property
    def accumulations(self):
        return self._metadata.accumulations

    @property
    def latitudes(self):
        return self._metadata.latitudes

    @property
    def longitudes(self):
        return self._metadata.longitudes

    @property
    def grid_points_mask(self):
        return self._metadata.grid_points_mask

    @cached_property
    def sources(self):
        return [SourceCheckpoint(self, _) for _ in self._metadata.sources(self.path)]

    def default_namer(self, *args, **kwargs):
        """Return a callable that can be used to name fields.
        In that case, return the namer that was used to create the
        training dataset.
        """
        return self._metadata.default_namer(*args, **kwargs)

    def report_error(self):
        self._metadata.report_error()

    def validate_environment(
        self,
        *,
        all_packages: bool = False,
        on_difference: str = "warn",
        exempt_packages: Optional[list[str]] = None,
    ) -> bool:
        return self._metadata.validate_environment(
            all_packages=all_packages, on_difference=on_difference, exempt_packages=exempt_packages
        )

    def open_dataset_args_kwargs(self, *, use_original_paths, from_dataloader=None):
        return self._metadata.open_dataset_args_kwargs(
            use_original_paths=use_original_paths,
            from_dataloader=from_dataloader,
        )

    def constant_forcings_inputs(self, runner, input_state):
        return self._metadata.constant_forcings_inputs(runner, input_state)

    def dynamic_forcings_inputs(self, runner, input_state):
        return self._metadata.dynamic_forcings_inputs(runner, input_state)

    def boundary_forcings_inputs(self, runner, input_state):
        return self._metadata.boundary_forcings_inputs(runner, input_state)

    def name_fields(self, fields, namer=None):
        return self._metadata.name_fields(fields, namer=namer)

    def sort_by_name(self, fields, namer=None, *args, **kwargs):
        return self._metadata.sort_by_name(fields, namer=namer, *args, **kwargs)

    def print_indices(self):
        return self._metadata.print_indices()

    def variable_categories(self):
        return self._metadata.variable_categories()

    def load_supporting_array(self, name):
        return self._metadata.load_supporting_array(name)

    @property
    def supporting_arrays(self):
        return self._metadata.supporting_arrays

    ###########################################################################

    @cached_property
    def lagged(self):
        """Return the list of timedelta for the `multi_step_input` fields."""
        result = list(range(0, self._metadata.multi_step_input))
        result = [-s * self._metadata.timestep for s in result]
        return sorted(result)

    @property
    def multi_step_input(self):
        return self._metadata.multi_step_input

    def print_variable_categories(self):
        return self._metadata.print_variable_categories()

    ###########################################################################
    # Data retrieval
    ###########################################################################

    def variables_from_input(self, *, include_forcings):
        return self._metadata.variables_from_input(include_forcings=include_forcings)

    @property
    def grid(self):
        return self._metadata.grid

    @property
    def area(self):
        return self._metadata.area

    def mars_by_levtype(self, levtype):
        return self._metadata.mars_by_levtype(levtype)

    def mars_requests(self, *, variables, dates, use_grib_paramid=False, always_split_time=False, **kwargs):
        from earthkit.data.utils.availability import Availability

        assert variables, "No variables provided"

        if not isinstance(dates, (list, tuple)):
            dates = [dates]

        dates = [to_datetime(d) for d in dates]

        assert dates, "No dates provided"

        result = []

        DEFAULT_KEYS = ("class", "expver", "type", "stream", "levtype")
        DEFAULT_KEYS_AND_TIME = ("class", "expver", "type", "stream", "levtype", "time")

        # The split oper/scda is a bit special
        KEYS = {("oper", "fc"): DEFAULT_KEYS_AND_TIME, ("scda", "fc"): DEFAULT_KEYS_AND_TIME}

        requests = defaultdict(list)

        for r in self._metadata.mars_requests(variables=variables, use_grib_paramid=use_grib_paramid):
            for date in dates:

                r = r.copy()

                base = date
                step = str(r.get("step", 0)).split("-")[-1]
                step = int(step)
                base = base - datetime.timedelta(hours=step)

                r["date"] = base.strftime("%Y-%m-%d")
                r["time"] = base.strftime("%H%M")

                r.update(kwargs)  # We do it here so that the Availability can use that information

                if always_split_time:
                    keys = DEFAULT_KEYS_AND_TIME
                else:
                    keys = KEYS.get((r.get("stream"), r.get("type")), DEFAULT_KEYS)
                key = tuple(r.get(k) for k in keys)

                # Special case because of oper/scda

                requests[key].append(r)

        result = []
        for reqs in requests.values():

            compressed = Availability(reqs)
            for r in compressed.iterate():
                for k, v in r.items():
                    if isinstance(v, (list, tuple)) and len(v) == 1:
                        r[k] = v[0]
                if r:
                    result.append(r)

        return result

    ###########################################################################
    # supporting arrays
    ###########################################################################

    @cached_property
    def _supporting_arrays(self):
        return self._metadata._supporting_arrays

    @property
    def name(self):
        return self._metadata.name

    ###########################################################################
    # Misc
    ###########################################################################

    def provenance_training(self):
        return self._metadata.provenance_training()


class SourceCheckpoint(Checkpoint):
    """A checkpoint that represents a source."""

    def __init__(self, owner, metadata):
        super().__init__(owner.path)
        self._owner = owner
        self._metadata = metadata

    def __repr__(self):
        return f"Source({self.name}@{self.path})"



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import earthkit.data as ekd

from ..decorators import main_argument
from . import input_registry
from .grib import GribInput

LOG = logging.getLogger(__name__)


@input_registry.register("grib")
@main_argument("path")
class GribFileInput(GribInput):
    """Handles grib files"""

    def __init__(self, context, path, *, namer=None, **kwargs):
        super().__init__(context, namer=namer, **kwargs)
        self.path = path

    def create_input_state(self, *, date):
        return self._create_input_state(ekd.from_source("file", self.path), variables=None, date=date)

    def load_forcings(self, *, variables, dates):
        return self._load_forcings(ekd.from_source("file", self.path), variables=variables, dates=dates)

    def template(self, variable, date, **kwargs):
        fields = ekd.from_source("file", self.path)
        data = self._find_variable(fields, variable)
        if len(data) == 0:
            return None
        return data[0]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


try:
    # NOTE: the `_version.py` file must not be present in the git repository
    #   as it is generated by setuptools at install time
    from ._version import __version__  # type: ignore
except ImportError:  # pragma: no cover
    # Local copy or not installed with setuptools
    __version__ = "999"



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import warnings
from collections import defaultdict

LOG = logging.getLogger(__name__)


def warn(func):
    def wrapper(*args, **kwargs):
        warnings.warn(f"Using legacy {func.__name__}, please try to patch your weights.")
        return func(*args, **kwargs)

    return wrapper


class LegacyMixin:

    # `self` is a `Metadata` object

    @warn
    def _legacy_variables_metadata(self):

        # Assumes ECMWF data from MARS
        result = {}
        unkowns = []
        for variable in self.variables:
            if variable in (
                "insolation",
                "cos_solar_zenith_angle",
                "cos_julian_day",
                "sin_julian_day",
                "cos_local_time",
                "sin_local_time",
            ):
                result[variable] = dict(computed_forcing=True, constant_in_time=False)
                continue

            if variable in ("cos_latitude", "cos_longitude", "sin_latitude", "sin_longitude"):
                result[variable] = dict(computed_forcing=True, constant_in_time=True)
                continue

            param_level = variable.split("_")
            if len(param_level) == 2:
                # Assumes levtype=pl for now
                levtype = "pl"

                try:
                    levelist = int(param_level[1])
                    mars = dict(param=param_level[0], levelist=levelist, levtype=levtype)
                    result[variable] = dict(mars=mars)
                    continue
                except ValueError:
                    pass

            if variable in (
                "z",
                "lsm",
                "sdor",
                "slor",
                "cl",
                "cvh",
                "cvl",
                "slt",
                "tvh",
                "tvl",
            ):
                mars = dict(param=variable, levtype="sfc")
                result[variable] = dict(mars=mars, constant_in_time=True)
                continue

            if variable in ("cp", "tp", "sf", "ro", "ssrd", "strd"):
                mars = dict(param=variable, levtype="sfc", type="fc", stream="oper")
                result[variable] = dict(mars=mars, process="accumulation", period=(0, 6))

                continue

            if variable in ("sp", "msl", "10u", "10v", "2t", "2d", "skt", "tcw"):
                mars = dict(param=variable, levtype="sfc")
                result[variable] = dict(mars=mars)
                continue

            unkowns.append(variable)
            mars = dict(param=variable, levtype="sfc")
            result[variable] = dict(mars=mars)

        if unkowns:
            warnings.warn(f"Unknown variables: {unkowns}, assuming an/sfc")

        return result

    def _legacy_check_variables_metadata(self, variables):

        if variables == {}:
            variables.update(self._legacy_variables_metadata())
            return

        first = True
        for variable, metadata in variables.items():
            if "mars" not in metadata:
                if "param" in metadata:
                    if first:
                        LOG.warning("Old metadata format detected. Please update your weights.")
                        first = False
                    metadata["mars"] = metadata.copy()

            if variable in (
                "insolation",
                "cos_solar_zenith_angle",
                "cos_julian_day",
                "sin_julian_day",
                "cos_local_time",
                "sin_local_time",
            ):
                ok = metadata.get("computed_forcing") is True and metadata.get("constant_in_time") is False
                if not ok:
                    if first:
                        LOG.warning("Old metadata format detected. Please update your weights.")
                        first = False
                    metadata["computed_forcing"] = True
                    metadata["constant_in_time"] = False

            if variable in ("cos_latitude", "cos_longitude", "sin_latitude", "sin_longitude"):
                ok = metadata.get("computed_forcing") is True and metadata.get("constant_in_time") is True
                if not ok:
                    if first:
                        LOG.warning("Old metadata format detected. Please update your weights.")
                        first = False
                    metadata["computed_forcing"] = True
                    metadata["constant_in_time"] = True

    @warn
    def _legacy_number_of_grid_points(self):

        POINTS = {"o96": 40_320, "n320": 542_080}

        return POINTS[self.grid.lower()]

    @warn
    def _legacy_data_request(self):
        from anemoi.utils.config import find

        result = find(self._metadata["dataset"], "data_request")
        if len(result) == 0:
            raise ValueError("No data_request found in metadata")

        if len(result) > 1:
            check = ("grid", "area")
            checks = defaultdict(set)
            for r in result:
                for c in check:
                    checks[c].add(str(r.get(c)))

            for c in check:
                if len(checks[c]) > 1:
                    warnings.warn(f"{c} is ambigous: {checks[c]}")

            result = [r for r in result if r["grid"] is not None]

        return result[0]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

LOG = logging.getLogger(__name__)


GRIB1_ONLY = []

GRIB2_ONLY = ["typeOfGeneratingProcess"]


def _param(param):
    try:
        int(param)
        return "paramId"
    except ValueError:
        try:
            float(param)
            return "param"
        except ValueError:
            return "shortName"


def grib_keys(
    *,
    values,
    template,
    accumulation,
    param,
    date,
    time,
    step,
    keys,
    quiet,
    grib1_keys={},
    grib2_keys={},
):
    result = keys.copy()

    edition = keys.get("edition")
    if edition is None and template is not None:
        edition = template.metadata("edition")
        # centre = template.metadata("centre")
        if edition == 2:
            productDefinitionTemplateNumber = template.metadata("productDefinitionTemplateNumber")
            if productDefinitionTemplateNumber in (8, 11) and not accumulation:
                if f"{param}-accumulation" not in quiet:
                    LOG.warning(
                        "%s: Template %s is accumulation but `accumulation` was not specified",
                        param,
                        productDefinitionTemplateNumber,
                    )
                    LOG.warning("%s: Setting `accumulation` to `True`", param)
                    quiet.add(f"{param}-accumulation")
                accumulation = True

    if edition is None:
        edition = 1

    if edition == 1:
        for k in GRIB2_ONLY:
            result.pop(k, None)

    if edition == 2:
        for k in GRIB1_ONLY:
            result.pop(k, None)

    result["edition"] = edition

    if param is not None:
        result.setdefault(_param(param), param)

        if edition == 1:
            result.update(grib1_keys.get(param, {}))

        if edition == 2:
            result.update(grib2_keys.get(param, {}))

    result.setdefault("type", "fc")

    if result.get("type") in ("an", "fc"):
        # For organisations that do not use type
        result.setdefault("dataType", result.pop("type"))

    # if stream is not None:
    #     result.setdefault("stream", stream)

    if date is not None:
        result["date"] = date

    if time is not None:
        result.setdefault("time", time)

    # 0: instantaneous, deterministic
    # 1: instantaneous, ensemble
    # 8: time processed, deterministic
    # 11: time processed, ensemble

    if accumulation:
        if edition == 1:
            result["step"] = step
        else:
            result["startStep"] = 0
            result["endStep"] = step
            result["stepType"] = "accum"

        if edition == 2:
            result["typeOfStatisticalProcessing"] = 1
            result["productDefinitionTemplateNumber"] = 8
            if result.get("type") in ("pf", "cf"):
                result["productDefinitionTemplateNumber"] = 11

    else:
        result["step"] = step
        if edition == 2:
            result["productDefinitionTemplateNumber"] = 0
            if result.get("type") in ("pf", "cf"):
                result["productDefinitionTemplateNumber"] = 1

    return result


def check_encoding(handle, keys):
    def same(w, v, k):
        if type(v) is type(w):
            return v == w
        return str(w) == str(v)

    mismatches = {}
    for k, v in keys.items():
        if k == "param":
            try:
                int(v)
                k = "paramId"
            except ValueError:
                try:
                    float(v)
                    k = "param"
                except ValueError:
                    k = "shortName"

        if k == "date":
            v = int(str(v).replace("-", ""))

        if k == "time":
            v = int(v)
            if v < 100:
                v *= 100

        if isinstance(v, int):
            w = handle.get_long(k)
        else:
            w = handle.get(k)

        if not same(w, v, k):
            mismatches[k] = 'Expected "{}" but got "{}"'.format(v, w)

    if mismatches:
        raise ValueError(f"GRIB field could not be encoded. Mismatches={mismatches}")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import ABC
from abc import abstractmethod

import earthkit.data as ekd
import numpy as np
from anemoi.transform.grids.unstructured import UnstructuredGridFieldList
from earthkit.data.indexing.fieldlist import FieldArray

from anemoi.inference.inputs.dataset import DatasetInput

LOG = logging.getLogger(__name__)


class Forcings(ABC):
    """Represents the forcings for the model."""

    def __init__(self, context):
        self.context = context
        self.checkpoint = context.checkpoint
        self.kinds = dict(unknown=True)  # Used for debugging

    @abstractmethod
    def load_forcings(self, state, date):
        pass

    def __repr__(self):
        return f"{self.__class__.__name__}"


class ComputedForcings(Forcings):
    """Compute forcings like `cos_julian_day` or `insolation`."""

    def __init__(self, context, variables, mask):
        super().__init__(context)
        self.variables = variables
        self.mask = mask
        self.kinds = dict(computed=True)  # Used for debugging

    def __repr__(self):
        return f"{self.__class__.__name__}({self.variables})"

    def load_forcings(self, state, dates):

        LOG.debug("Adding dynamic forcings %s", self.variables)

        if not isinstance(dates, (list, tuple)):
            dates = [dates]

        source = UnstructuredGridFieldList.from_values(
            latitudes=state["latitudes"],
            longitudes=state["longitudes"],
        )

        ds = ekd.from_source("forcings", source, date=dates, param=self.variables)

        assert len(ds) == len(self.variables) * len(dates), (len(ds), len(self.variables), dates)

        def rename(f, _, metadata):
            return metadata["param"]

        ds = FieldArray([f.clone(name=rename) for f in ds])

        forcing = ds.order_by(name=self.variables, valid_datetime="ascending")

        # Forcing are sorted by `compute_forcings`  in the order (varaible, date)

        return forcing.to_numpy(dtype=np.float32, flatten=True).reshape(len(self.variables), len(dates), -1)


class CoupledForcings(Forcings):
    """Retrieve forcings from the input."""

    def __init__(self, context, input, variables, mask):
        super().__init__(context)
        self.variables = variables
        self.mask = mask
        self.input = input
        # self.grid = context.checkpoint.grid
        # self.area = context.checkpoint.area
        # self.use_grib_paramid = True  # TODO: find a way to `use_grib_paramid``
        self.kinds = dict(retrieved=True)  # Used for debugging

    def __repr__(self):
        return f"{self.__class__.__name__}({self.variables})"

    def load_forcings(self, state, dates):
        data = self.input.load_forcings(variables=self.variables, dates=dates)

        expected_shape = (len(self.variables), len(dates), state["latitudes"].size)
        assert data.shape == expected_shape, (data.shape, expected_shape)

        return data

        # assert False, "Not implemented yet"
        # from .inputs.mars import retrieve

        # requests = self.context.checkpoint.mars_requests(
        #     variables=self.variables,
        #     dates=dates,
        #     use_grib_paramid=self.use_grib_paramid,
        # )

        # if not requests:
        #     raise ValueError("No requests for %s (%s)" % (self.variables, dates))

        # for r in requests:
        #     LOG.info("Request: %s", r)

        # fields = retrieve(requests=requests, grid=self.grid, area=self.area, expver=1)

        # if not fields:
        #     raise ValueError("No fields retrieved for {self.variables} ({dates})")

        # fields = self.checkpoint.name_fields(fields).order_by(name=self.variables, valid_datetime="ascending")

        # return fields.to_numpy(dtype=np.float32, flatten=True).reshape(len(self.variables), len(dates), -1)


class BoundaryForcings(Forcings):
    """Retrieve boundary forcings from the input."""

    def __init__(self, context, input, variables, variables_mask):
        super().__init__(context)
        self.variables = variables
        self.variables_mask = variables_mask
        assert isinstance(input, DatasetInput), "Currently only boundary forcings from dataset supported."
        self.input = input
        if "output_mask" in context.checkpoint._supporting_arrays:
            self.spatial_mask = ~context.checkpoint.load_supporting_array("output_mask")
        else:
            self.spatial_mask = np.array([False] * len(input["latitudes"]), dtype=bool)
        self.kinds = dict(retrieved=True)  # Used for debugging

    def __repr__(self):
        return f"{self.__class__.__name__}({self.variables})"

    def load_forcings(self, state, dates):
        data = self.input.load_forcings(variables=self.variables, dates=dates)
        data = data[..., self.spatial_mask]

        expected_shape = (len(self.variables), len(dates), state["latitudes"][self.spatial_mask].size)
        assert data.shape == expected_shape, (data.shape, expected_shape)

        return data



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
from abc import ABC
from abc import abstractmethod
from functools import cached_property

LOG = logging.getLogger(__name__)


class Output(ABC):
    """_summary_"""

    def __init__(self, context, output_frequency=None, write_initial_state=None):

        self.context = context
        self.checkpoint = context.checkpoint
        self.reference_date = None

        self._write_step_zero = write_initial_state
        self._output_frequency = output_frequency

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def step(self, state):
        return state["date"] - self.reference_date

    def write_initial_state(self, state):
        self._init(state)
        if self.write_step_zero:
            return self.write_initial_step(state)

    def write_state(self, state):
        self._init(state)

        step = self.step(state)
        if self.output_frequency is not None:
            if (step % self.output_frequency).total_seconds() != 0:
                return

        return self.write_step(state)

    def _init(self, state):
        if self.reference_date is not None:
            return

        self.reference_date = state["date"]

        self.open(state)

    def write_initial_step(self, state):
        """This method should not be called directly
        call `write_initial_state` instead.
        """
        reduced_state = self.reduce(state)
        self.write_step(reduced_state)

    @abstractmethod
    def write_step(self, state):
        """This method should not be called directly
        call `write_state` instead.
        """
        pass

    def reduce(self, state):
        """Creates new state which is projection of original state on the last step in the multi-steps dimension."""
        reduced_state = state.copy()
        reduced_state["fields"] = {}
        for field, values in state["fields"].items():
            reduced_state["fields"][field] = values[-1, :]
        return reduced_state

    def open(self, state):
        # Override this method when initialisation is needed
        pass

    def close(self):
        pass

    @cached_property
    def write_step_zero(self):
        if self._write_step_zero is not None:
            return self._write_step_zero

        return self.context.write_initial_state

    @cached_property
    def output_frequency(self):
        from anemoi.utils.dates import as_timedelta

        if self._output_frequency is not None:
            return as_timedelta(self._output_frequency)

        if self.context.output_frequency is not None:
            return as_timedelta(self.context.output_frequency)

        return None

    def print_summary(self, depth=0):
        LOG.info(
            "%s%s: output_frequency=%s write_initial_state=%s",
            " " * depth,
            self,
            self.output_frequency,
            self.write_step_zero,
        )


class ForwardOutput(Output):
    """Subclass of Output that forwards calls to other outputs
    Subclass from that class to implement the desired behaviour of `output_frequency`
    which should only apply to leaves
    """

    def __init__(self, context, output_frequency=None, write_initial_state=None):
        super().__init__(context, output_frequency=None, write_initial_state=write_initial_state)
        if self.context.output_frequency is not None:
            LOG.warning("output_frequency is ignored for '%s'", self.__class__.__name__)

    @cached_property
    def output_frequency(self):
        return None



import datetime

from anemoi.inference.inputs.gribfile import GribFileInput
from anemoi.inference.outputs.gribfile import GribFileOutput
from anemoi.inference.runners import DefaultRunner

# Create a runner with the checkpoint file
runner = DefaultRunner("checkpoint.ckpt")

# Select a starting date
date = datetime.datetime(2024, 10, 25)

input = GribFileInput(runner, "input.grib")
output = GribFileOutput(runner, "output.grib")

input_state = input.create_input_state(date)

# Write the initial state to the output file
output.write_initial_state(input_state)

# Run the model and write the output to the file

for state in runner.run(input_state=input_state, lead_time=240):
    output.write_state(state)

# Close the output file
output.close()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from typing import Any
from typing import List

from anemoi.utils.humanize import plural

LOG = logging.getLogger(__name__)


def check_data(title: str, data: Any, variables: List[str], dates: List[datetime.datetime]) -> None:
    expected = len(variables) * len(dates)

    if len(data) != expected:

        from anemoi.utils.text import table

        LOG.error("Data check failed for %s", title)

        nvars = plural(len(variables), "variable")
        ndates = plural(len(dates), "date")
        nfields = plural(expected, "field")
        msg = f"Expected ({nvars}) x ({ndates}) = {nfields}, got {len(data)}"
        LOG.error("%s", msg)

        cols = {}
        rows = {}
        t = []
        for i, d in enumerate(sorted(dates)):
            cols[d.isoformat()] = i + 1

        for i in range(len(variables)):
            name = variables[i]
            while len(t) <= i:
                t.append([name] + (["❌"] * len(cols)))

            t[i][0] = name
            rows[name] = i

        for field in data:
            name, date = field.metadata("name"), field.metadata("valid_datetime")
            if t[rows[name]][cols[date]] == "❌":
                t[rows[name]][cols[date]] = ""
            t[rows[name]][cols[date]] += "✅"

        print(table(t, ["name"] + [_.isoformat() for _ in sorted(dates)], "<||"))

        raise ValueError(msg)

    assert len(data) == len(variables) * len(dates)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import base64
import logging
import zlib

import earthkit.data as ekd

from ..input import Input
from . import input_registry

LOG = logging.getLogger(__name__)

# 1 - Get a GRIB with mars: retrieve,param=tp,levtype=sfc,type=fc,step=6,target=data.grib,grid=0.25/0.25
# 2 - grib_set -s edition=2,packingType=grid_ccsds data.grib data.grib2
# 3 - grib_set -d 0 data.grib2 out.grib
# 4 - python -c 'import base64, sys, zlib;print(base64.b64encode(zlib.compress(open(sys.argv[1], "rb").read())))' out.grib

# tp in grib2, 0.25/0.25 grid
TEMPLATE_0p25 = """
eJxzD/J0+v+fgYkBAh4AsSgjQxIDgyIjI/sLbnEeoAAjEAsyASlGBk4WRgMDA0OggAczAwP/9QVgXf9RAQMDK1Cc6SJMijXWqw
HENmgFMkQLIwsYmC9MAGMgsGIBEhyMBxn+zwMqBVnFxohkFsgNQhAnMDAygan/ULdKsoIdoAViNwiDSD4FBpBFbGwgJazs5kAA
AKlBO70=
"""

TEMPLATE_N320 = """
eJztlDFLw0AYht87m1ZRCgpFCg4VHEQs1C6KOASpqDjVSRykiIJDBzddhAyKji46SsE/0MHd6uTg4KiTuoiTODh1ie93SfXaSO
gP8B5yd4TjI1y+91lcXZ73fWiY4bxwyihsAqNKpd77hwf4QvEZ1FwU+hKqUChM8eRSD9C74GmFcfjRAWhPKsreWdd5U/6Im8zu
80ZwQLnAELIYY4U8ipiDiyWygjLWUcEWqtgnBzjGKTlHDVeGBrnDo+EVn4YmPy9NcoZJ4hpKasdij3jklJypesit4eWHN/UlaF
6LHvnBtVgL8do4IZchDxGe9EdAD0JyFhOk8ieHHTRiuCfv7SRgMRPDdgxehHoXXNs4+Kcb2m+tq3uO/p24fxnXA1avdPaR6a1G
DJ19+nc3S5/bfR9moZUNpiSanFamJF/tiWvl0M7mb2KZ3yDJTPRvuoO0t7J/ZlzgEXGD7YpSaBBxSWAV8QvQDI3zGhrozvgoMF
ONlhJXHdNa4q4qLVahzcq0mrjNpeWKtN04rZel/YDZBKdedQP/glYU1SaV5VFx8FCgYChtFj9wNbKOCBgTsvekFNI5iHSTSTni
pKY5vgHl2Dri
"""

TEMPLATE_O96 = """
eJxlkjFIW1EUhv//mailoVSNaCFDChlEMjwhoEgGkYAWMohkcAgllEAdgjhkcMiQ4RUcMjg4ONji4ODg4ODg4KDg4ODQwaGDQw
cHB4cODh2E1+/FDoXex7v3cO6555z/u3dp9cNiHCtQfwQNpnHrk/TeHnp4PZHBYf6RgMV6lXIYhjP4rgfwfu0G1pTi/webF0nG
xE6vfoz66b9gjH/eaPwNoFhWk8oprwJZigpV0pzKWlBFy6pqRTWtqU5gU+tqaVNtbamjriJtq6cd7WpP+zrQoY50rBOd6kznVL
7StW70Xbf6oTv91L0e9KhfetJvPSMk5WFn/NZZTzrnvAuectGhS55z2QuueNlVr7jmNdfdcNPrbnnTbW+5464jb7vnHb4eVoSn
w06biBaRTU7UOVkjQ5VMFTKWyVyiQpFKBSrmqJylgwydpCD8TGdPdPhIp/d0fEfntyi4QckVis5RdorCY5Qeongf5bsQ6EEigk
gHMm0ItSDVhFgdcjUIViFZgWgZsiUIFyFdgHgO8lkuYT7FNOxLxd+4leSqB/3PPSZvYPTlCchBf4lf3orepZMHoOnE7o4l85s8
jZBgMAlJD80y/gBx3pOi
"""

TEMPLATES = {(0.25, 0.25): TEMPLATE_0p25, "N320": TEMPLATE_N320, "O96": TEMPLATE_O96}


@input_registry.register("templates")
class TemplatesInput(Input):
    """A dummy input that creates GRIB templates"""

    def __init__(self, context, *args, **kwargs):
        super().__init__(context)

    def __repr__(self):
        return f"TemplatesInput({self.context.checkpoint.grid})"

    def create_input_state(self, *, date):
        raise NotImplementedError("TemplatesInput.create_input_state() not implemented")

    def template(self, variable, date, **kwargs):

        # import eccodes
        typed = self.context.checkpoint.typed_variables[variable]

        if not typed.is_accumulation:
            return None

        grid = self.context.checkpoint.grid
        if isinstance(grid, str):
            grid = grid.upper()

        template = zlib.decompress(base64.b64decode(TEMPLATES[grid]))
        # eccodes.codes_new_from_message(template)

        return ekd.from_source("memory", template)[0]

    def load_forcings(self, variables, dates):
        raise NotImplementedError("TemplatesInput.load_forcings() not implemented")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from __future__ import annotations

import itertools
import re
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Optional
from typing import TypeVar

import numpy as np

if TYPE_CHECKING:
    import earthkit.data as ekd
    import xarray as xr

Array = TypeVar("Array")


def extract_keys(s: str) -> list[str]:
    # Use regular expression to find all occurrences of {key}
    keys = re.findall(r"\{(.*?)\}", s)
    return keys


def permute_dict(d: dict) -> list[dict]:
    # Get the keys and values from the dictionary
    keys = d.keys()
    values = d.values()

    # Generate all permutations using itertools.product
    permutations = itertools.product(*values)

    # Convert permutations into a list of dictionaries
    result = [dict(zip(keys, permutation)) for permutation in permutations]

    return result


def summarise_list(lst: list, max_length: int) -> str:
    if len(lst) > max_length:
        summary = f"{str(lst[:3])[:-1]} ... {str(lst[-3:])[1:]} (Total: {len(lst)} items)"
        return summary
    return str(lst)


class State(dict[str, np.ndarray]):
    """A collection of data for inference."""

    def __init__(self, data: dict[str, np.ndarray] = None, *, private_info: Any = None, **kwargs):
        """Create a State object.

        Parameters
        ----------
        data : dict[str, np.ndarray]
            Dictionary of data to store in the State
        private_info : Any, optional
            Private info to pass with the State, by default None
        """
        super().__init__(data or {}, **kwargs)
        self.__private_info = private_info

    def __repr__(self):
        return f"State({summarise_list(list(self.keys()), 8)}, private_info = {self.__private_info})"

    def take(self, axis: int, indices: int | slice | tuple[int, ...], *, copy: bool = True) -> State:
        """Take data from the State with a specific dimension and index.

        Parameters
        ----------
        axis : int
            Dimension to take data from
        indices : int | slice | tuple[int, ...]
            Index to take data with
        copy : bool, optional
            Whether to copy the data, If False, the data will be taken without copying,
            by default True

        Returns
        -------
        State
            State object with the dimension and index taken on all keys

        Examples
        --------
        >>> data = np.random.rand(3, 2, 4)
        >>> names = ["a", "b", "c"]
        >>> state = State.from_numpy(data, names)
        >>> state.take(0, 0)
        State(['a', 'b', 'c'], private_info = None)
        >>> state.take(0, 0).shape
        {'a': (2, 4), 'b': (2, 4), 'c': (2, 4)}
        """
        if copy:
            return State(
                {key: value.take(indices, axis) for key, value in self.items()}, private_info=self.__private_info
            )

        index = [[slice(None)] * axis, indices, [slice(None)] * (len(self[list(self.keys())[0]].shape) - axis - 1)]
        index = [i for i in index if (isinstance(i, list) and len(i) > 1) or not isinstance(i, list)]
        return State({key: value.__getitem__(*index) for key, value in self.items()}, private_info=self.__private_info)

    def to_array(
        self,
        order: list[str],
        *,
        stack_function: Callable[[list[Array], Any], Array] = np.stack,
        array_function: Callable[[np.array], Array] = np.array,
        **kwargs,
    ) -> Array:
        """Convert the State to an array.

        Parameters
        ----------
        order : list[str]
            Order to extract the keys from the State
        stack_function : Callable, optional
            Function to stack arrays with, by default np.stack
        array_function: Callable, optional
            Function to convert arrays with, must take np.array, by default np.array
        **kwargs:
            Additional keyword arguments to pass to the stack_function

        Returns
        -------
        Array
            Stacked array

        Raises
        ------
        ValueError
            If any keys in order are not in the State

        Examples
        --------
        >>> import numpy as np
        >>> data = np.random.rand(3, 2, 4)
        >>> names = ["a", "b", "c"]
        >>> state = State.from_numpy(data, names)
        >>> state.to_array(names[::-1]).shape
        (3, 2, 4)
        """
        if any(key not in self for key in order):
            raise ValueError("Some keys in order are not in the State", self.keys(), order)
        return stack_function([array_function(self.get(key)) for key in order], **kwargs)

    @property
    def shape(self) -> dict[str, tuple[int, ...]]:
        """Get the shape of the State"""
        return {key: value.shape for key, value in self.items()}

    @classmethod
    def from_xarray(
        self,
        data: xr.Dataset | xr.DataArray,
        *,
        flatten: Optional[str] = None,
        variable_dim: str = "variable",
        private_info: Any = None,
    ) -> State:
        """Convert an xarray Dataset or DataArray to a State.

        Parameters
        ----------
        data : xr.Dataset | xr.DataArray
            xr.Dataset or xr.DataArray object to convert to a State
        flatten : str | None, optional
            F-string dictating how to flatten dimensions, by default None
            E.g. "{variable}_{level}" will create a new key for each level value
        variable_dim : str, optional
            Dimension name for variables if xr.DataArray given, by default "variable"
        private_info : Any, optional
            Private info to pass to State, by default None

        Returns
        -------
        State
            State object

        Raises
        ------
        ValueError
            If no keys are found in flatten
        KeyError
            If variable_dim is not found in keys

        Examples
        --------
        >>> ds = xr.tutorial.load_dataset("air_temperature")
        >>> ds
        <xarray.Dataset> Size: 31MB
        Dimensions:  (lat: 25, time: 2920, lon: 53)
        Coordinates:
        * lat      (lat) float32 100B 75.0 72.5 70.0 67.5 65.0 ... 22.5 20.0 17.5 15.0
        * lon      (lon) float32 212B 200.0 202.5 205.0 207.5 ... 325.0 327.5 330.0
        * time     (time) datetime64[ns] 23kB 2013-01-01 ... 2014-12-31T18:00:00
        Data variables:
            air      (time, lat, lon) float64 31MB ...
        Attributes:
            Conventions:  COARDS
            title:        4x daily NMC reanalysis (1948)
            description:  Data is from NMC initialized reanalysis\n(4x/day).  These a...
            platform:     Model
            references:   http://www.esrl.noaa.gov/psd/data/gridded/data.ncep.reanaly...
        >>> State.from_xarray(ds))
        State(['air'], private_info = None)
        >>> State.from_xarray(ds, flatten="{variable}_{lat}"))
        State(['air_75.0', 'air_72.5', 'air_70.0', 'air_67.5', 'air_65.0', 'air_62.5', 'air_60.0', 'air_57.5', 'air_55.0', 'air_52.5', 'air_50.0', 'air_47.5', 'air_45.0', 'air_42.5', 'air_40.0', 'air_37.5', 'air_35.0', 'air_32.5', 'air_30.0', 'air_27.5', 'air_25.0', 'air_22.5', 'air_20.0', 'air_17.5', 'air_15.0'], private_info = None)
        """
        import xarray as xr

        # Flatten the data if required
        if flatten is not None:
            keys = extract_keys(flatten)
            if len(keys) == 0:
                raise ValueError("No keys found in flatten")

            if variable_dim not in keys:
                raise KeyError(f"variable_dim {variable_dim} not found in keys {keys}")

            new_data = xr.Dataset()
            for var in data.coords[variable_dim].values:
                var_ds = data.sel({variable_dim: var})

                if not any(key in var_ds.dims for key in keys):
                    new_data[var] = data.sel({variable_dim: var})

                for perm in permute_dict({key: list(np.atleast_1d(var_ds.coords[key].values)) for key in keys}):
                    new_data[flatten.format(**perm)] = data.sel({variable_dim: var, **perm})

            data = new_data

        # Convert to DataArray if Dataset so dimensions are in the correct order
        if isinstance(data, xr.Dataset):
            data = data.to_dataarray(dim=variable_dim)

        dims = list(data.dims)
        dims.remove(variable_dim)
        data = data.transpose(variable_dim, *dims)

        # Get all variables and their values in the data
        variable_dict = {}
        for var in data.coords[variable_dim].values:
            variable_dict[var] = data.sel({variable_dim: var}).values
        return State(variable_dict, private_info=private_info)

    @classmethod
    def from_earthkit(self, fieldlist: "ekd.FieldList", private_info: Any = None, **kwargs) -> State:
        """Convert a FieldList to a State.

        Parameters
        ----------
        fieldlist : ekd.FieldList
            earthkit data FieldList object
        private_info : Any
            Private information to store in the State object
        **kwargs:
            Additional keyword arguments to pass to the to_xarray method of the FieldList object.
            See /earthkit/data/utils/xarray/engine.py/EarthkitBackendEntrypoint/open_dataset for more information

        Returns
        -------
        State
            State object

        Examples
        --------
        >>> import earthkit.data as ekd
        >>> ekd.download_example_file("test6.grib")
        >>> fieldlist = ekd.from_source("file", "test6.grib")
        >>> fieldlist
        GRIBReader(test6.grib)
        >>> State.from_earthkit(fieldlist)
        State(['t', 'u', 'v'], private_info = None)
        >>> State.from_earthkit(fieldlist, variable_key="par_lev_type", remapping={"par_lev_type": "{param}_{levelist}"})
        State(['t_1000', 't_850', 'u_1000', 'u_850', 'v_1000', 'v_850'], private_info = None)
        """
        return self.from_xarray(fieldlist.to_xarray(**kwargs), private_info=private_info)

    @classmethod
    def from_numpy(
        self,
        data: np.ndarray,
        names: list[str],
        *,
        axis: int = 0,
        private_info: Any = None,
    ) -> State:
        """Convert a numpy array to a State.

        Parameters
        ----------
        data : np.ndarray
            Numpy array to convert to a State
        names : list[str]
            Names upon `axis` to use as keys
        axis : int, optional
            Axis to split data upon, by default 0
        private_info : Any, optional
            Private information to store in the State object, by default None

        Returns
        -------
        State
            State object

        Examples
        --------
        >>> import numpy as np
        >>> data = np.random.rand(3, 2, 4)
        >>> names = ["a", "b", "c"]
        >>> State.from_numpy(data, names)
        State(['a', 'b', 'c'], private_info = None)
        >>> names = ["a", "b", "c", "d"]
        >> State.from_numpy(data, names, axis=2)
        State(['a', 'b', 'c', 'd'], private_info = None)
        """

        if axis != 0:
            data = np.moveaxis(data, axis, 0)
        return State(dict(zip(names, data)), private_info=private_info)

    def copy(self):
        """Copy the State object."""
        return State(dict(self.items()), private_info=self.__private_info)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from copy import deepcopy

from . import Command

LOG = logging.getLogger(__name__)


class PatchCmd(Command):
    """Patch a checkpoint file."""

    need_logging = True
    _cache = {}

    def add_arguments(self, command_parser):
        command_parser.add_argument("path", help="Path to the checkpoint.")
        command_parser.add_argument("--sanitise", action="store_true", help="Sanitise the metadata.")
        command_parser.add_argument("--force", action="store_true", help="Force the patching.")

    def run(self, args):
        from anemoi.utils.checkpoints import load_metadata
        from anemoi.utils.checkpoints import metadata_root
        from anemoi.utils.checkpoints import replace_metadata

        from anemoi.inference.metadata import Metadata

        root = metadata_root(args.path)

        original_metadata, supporting_arrays = load_metadata(args.path, supporting_arrays=True)
        metadata = deepcopy(original_metadata)

        # Patch the metadata
        while True:
            previous = deepcopy(metadata)
            metadata, supporting_arrays = Metadata(metadata).patch_metadata(supporting_arrays, root, force=args.force)
            if metadata == previous:
                break
            LOG.info("Metadata patched")

        if metadata != original_metadata:
            LOG.info("Patching metadata")
            assert "sources" in metadata["dataset"]
            replace_metadata(args.path, metadata, supporting_arrays)


command = PatchCmd



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging

from ..config import load_config
from ..runners import create_runner
from . import Command

LOG = logging.getLogger(__name__)


def _run(runner, config):
    input = runner.create_input()
    output = runner.create_output()

    input_state = input.create_input_state(date=config.date)

    output.write_initial_state(input_state)

    for state in runner.run(input_state=input_state, lead_time=config.lead_time):
        output.write_state(state)

    output.close()


class RunCmd(Command):
    """Run inference from a config yaml file."""

    need_logging = False

    def add_arguments(self, command_parser):
        command_parser.add_argument("--defaults", action="append", help="Sources of default values.")
        command_parser.add_argument("config", help="Path to config file.")
        command_parser.add_argument("overrides", nargs="*", help="Overrides.")

    def run(self, args):

        config = load_config(args.config, args.overrides, defaults=args.defaults)

        if config.description is not None:
            LOG.info("%s", config.description)

        runner = create_runner(config)

        _run(runner, config)


command = RunCmd



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from unittest.mock import patch

import pytest

import anemoi.inference.checkpoint
from anemoi.inference.runner import Runner

from ..metadata.fake_metadata import FakeMetadata


@pytest.fixture(scope="session")
def fake_huggingface_repo(tmp_path_factory):
    """Create a fake huggingface repo download"""
    tmp_dir = tmp_path_factory.mktemp("repo")
    fn = tmp_dir / "model.ckpt"
    fn.write_text("TESTING", encoding="utf-8")
    return tmp_dir


@pytest.fixture(scope="session")
def fake_huggingface_ckpt(tmp_path_factory):
    """Create a fake huggingface ckpt download"""
    tmp_dir = tmp_path_factory.mktemp("repo")
    fn = tmp_dir / "model.ckpt"
    fn.write_text("TESTING", encoding="utf-8")
    return fn


@patch("huggingface_hub.snapshot_download")
@pytest.mark.parametrize("ckpt", ["organisation/test_repo"])
def test_huggingface_repo_download_str(huggingface_mock, monkeypatch, ckpt, fake_huggingface_repo):

    monkeypatch.setattr(anemoi.inference.checkpoint.Checkpoint, "_metadata", FakeMetadata())
    huggingface_mock.return_value = fake_huggingface_repo

    runner = Runner({"huggingface": ckpt})
    assert runner.checkpoint.path == str(fake_huggingface_repo / "model.ckpt")

    assert huggingface_mock.called
    huggingface_mock.assert_called_once_with(repo_id=ckpt)


@patch("huggingface_hub.snapshot_download")
@pytest.mark.parametrize("ckpt", [{"repo_id": "organisation/test_repo"}])
def test_huggingface_repo_download_dict(huggingface_mock, monkeypatch, ckpt, fake_huggingface_repo):

    monkeypatch.setattr(anemoi.inference.checkpoint.Checkpoint, "_metadata", FakeMetadata())
    huggingface_mock.return_value = fake_huggingface_repo

    runner = Runner({"huggingface": ckpt})
    assert runner.checkpoint.path == str(fake_huggingface_repo / "model.ckpt")

    assert huggingface_mock.called
    huggingface_mock.assert_called_once_with(**ckpt)


@patch("huggingface_hub.hf_hub_download")
@pytest.mark.parametrize("ckpt", [{"repo_id": "organisation/test_repo", "filename": "model.ckpt"}])
def test_huggingface_file_download(huggingface_mock, monkeypatch, ckpt, fake_huggingface_ckpt):

    monkeypatch.setattr(anemoi.inference.checkpoint.Checkpoint, "_metadata", FakeMetadata())
    huggingface_mock.return_value = fake_huggingface_ckpt

    runner = Runner({"huggingface": ckpt})
    assert runner.checkpoint.path == str(fake_huggingface_ckpt)

    assert huggingface_mock.called
    huggingface_mock.assert_called_once_with(**ckpt)



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

from . import runner_registry
from .default import DefaultRunner

LOG = logging.getLogger(__name__)


@runner_registry.register("crps")
class CrpsRunner(DefaultRunner):
    def predict_step(self, model, input_tensor_torch, fcstep, **kwargs):
        return model.predict_step(input_tensor_torch, fcstep=fcstep)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os

import numpy as np

from ..output import Output
from . import output_registry

LOG = logging.getLogger(__name__)


def fix(lons):
    return np.where(lons > 180, lons - 360, lons)


@output_registry.register("plot")
class PlotOutput(Output):
    """_summary_"""

    def __init__(
        self,
        context,
        path,
        variables=all,
        strftime="%Y%m%d%H%M%S",
        template="plot_{variable}_{date}.{format}",
        dpi=300,
        format="png",
        output_frequency=None,
        write_initial_state=None,
    ):
        super().__init__(context, output_frequency=output_frequency, write_initial_state=write_initial_state)
        self.path = path
        self.format = format
        self.variables = variables
        self.strftime = strftime
        self.template = template
        self.dpi = dpi

        if self.variables is not all:
            if not isinstance(self.variables, (list, tuple)):
                self.variables = [self.variables]

    def write_step(self, state):
        import cartopy.crs as ccrs
        import cartopy.feature as cfeature
        import matplotlib.pyplot as plt
        import matplotlib.tri as tri

        os.makedirs(self.path, exist_ok=True)

        last_missing = None

        for name, values in state["fields"].items():

            if self.variables is not all and name not in self.variables:
                continue

            _, ax = plt.subplots(subplot_kw={"projection": ccrs.PlateCarree()})
            ax.coastlines()
            ax.add_feature(cfeature.BORDERS, linestyle=":")

            missing_values = np.isnan(values)

            values = values[~missing_values]
            if last_missing is None or np.any(last_missing != missing_values):
                longitudes = state["longitudes"][~missing_values]
                latitudes = state["latitudes"][~missing_values]
                triangulation = tri.Triangulation(fix(longitudes), latitudes)
                last_missing = missing_values

            _ = ax.tricontourf(triangulation, values, levels=10, transform=ccrs.PlateCarree())

            ax.tricontour(
                triangulation,
                values,
                levels=10,
                colors="black",
                linewidths=0.5,
                transform=ccrs.PlateCarree(),
            )

            date = state["date"].strftime("%Y-%m-%d %H:%M:%S")
            ax.set_title(f"{name} at {date}")

            date = state["date"].strftime(self.strftime)
            fname = self.template.format(date=date, variable=name, format=self.format)
            fname = os.path.join(self.path, fname)

            plt.savefig(fname, dpi=self.dpi, bbox_inches="tight")
            plt.close()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import re
from collections import defaultdict

import numpy as np
from earthkit.data.indexing.fieldlist import FieldArray
from earthkit.data.utils.dates import to_datetime

from ..checks import check_data
from ..input import Input

LOG = logging.getLogger(__name__)


class NoMask:
    """No mask to apply"""

    def apply(self, field):
        return field


class ApplyMask:
    """Apply a mask to a field"""

    def __init__(self, mask):
        self.mask = mask

    def apply(self, field):
        return field[self.mask]


class RulesNamer:
    """A namer that uses rules to generate names"""

    def __init__(self, rules, default_namer):
        self.rules = rules
        self.default_namer = default_namer

    def __call__(self, field, original_metadata):
        for rule in self.rules:
            assert len(rule) == 2, rule
            ok = True
            for k, v in rule[0].items():
                if original_metadata.get(k) != v:
                    ok = False
            if ok:
                return self.substitute(rule[1], field, original_metadata)

        return self.default_namer(field, original_metadata)

    def substitute(self, template, field, original_metadata):
        matches = re.findall(r"\{(.+?)\}", template)
        matches = {m: original_metadata.get(m) for m in matches}
        return template.format(**matches)


class EkdInput(Input):
    """Handles earthkit-data FieldList as input"""

    def __init__(self, context, *, namer=None):
        super().__init__(context)

        if isinstance(namer, dict):
            # TODO: a factory for namers
            assert "rules" in namer, namer
            assert len(namer) == 1, namer
            namer = RulesNamer(namer["rules"], self.checkpoint.default_namer())

        self._namer = namer if namer is not None else self.checkpoint.default_namer()
        assert callable(self._namer), type(self._namer)

    def _create_input_state(
        self,
        input_fields,
        *,
        variables=None,
        date=None,
        latitudes=None,
        longitudes=None,
        dtype=np.float32,
        flatten=True,
    ):

        if variables is None:
            variables = self.checkpoint.variables_from_input(include_forcings=True)

        if len(input_fields) == 0:
            raise ValueError("No input fields provided")

        # Newer checkpoints may have latitudes and longitudes
        if latitudes is None:
            latitudes = self.checkpoint.latitudes
            if latitudes is not None:
                LOG.info(
                    "%s: using `latitudes` found in the checkpoint.",
                    self.__class__.__name__,
                )

        if longitudes is None:
            longitudes = self.checkpoint.longitudes
            if longitudes is not None:
                LOG.info(
                    "%s: using `longitudes` found in the checkpoint.git.",
                    self.__class__.__name__,
                )

        if date is None:
            date = input_fields.order_by(valid_datetime="ascending")[-1].datetime()["valid_time"]
            LOG.info(
                "%s: `date` not provided, using the most recent date: %s", self.__class__.__name__, date.isoformat()
            )

        date = to_datetime(date)
        dates = [date + h for h in self.checkpoint.lagged]
        date_to_index = {d.isoformat(): i for i, d in enumerate(dates)}

        input_state = dict(date=date, latitudes=latitudes, longitudes=longitudes, fields=dict())

        fields = input_state["fields"]

        input_fields = self._filter_and_sort(input_fields, variables=variables, dates=dates, title="Create input state")
        mask = self.checkpoint.grid_points_mask
        mask = ApplyMask(mask) if mask is not None else NoMask()

        check = defaultdict(set)

        for field in input_fields:

            if input_state["latitudes"] is None:
                input_state["latitudes"], input_state["longitudes"] = field.grid_points()
                LOG.info(
                    "%s: using `latitudes` and `longitudes` from the first input field",
                    self.__class__.__name__,
                )

            name, valid_datetime = field.metadata("name"), field.metadata("valid_datetime")
            if name not in fields:
                fields[name] = np.full(
                    shape=(len(dates), self.checkpoint.number_of_grid_points),
                    fill_value=np.nan,
                    dtype=dtype,
                )

            date_idx = date_to_index[valid_datetime]

            try:
                fields[name][date_idx] = mask.apply(field.to_numpy(dtype=dtype, flatten=flatten))
            except ValueError:
                LOG.error("Error with field %s: expected shape=%s, got shape=%s", name, fields[name].shape, field.shape)
                LOG.error("dates %s", dates)
                LOG.error("number_of_grid_points %s", self.checkpoint.number_of_grid_points)
                raise

            if date_idx in check[name]:
                LOG.error("Duplicate dates for %s: %s", name, date_idx)
                LOG.error("Expected %s", list(date_to_index.keys()))
                LOG.error("Got %s", list(check[name]))
                raise ValueError(f"Duplicate dates for {name}")

            check[name].add(date_idx)

        for name, idx in check.items():
            if len(idx) != len(dates):
                LOG.error("Missing dates for %s: %s", name, idx)
                LOG.error("Expected %s", list(date_to_index.keys()))
                LOG.error("Got %s", list(idx))
                raise ValueError(f"Missing dates for {name}")

        # This is our chance to communicate output object
        # This is useful for GRIB that requires a template field
        # to be used as output
        self.set_private_attributes(input_state, input_fields)

        return input_state

    def _filter_and_sort(self, data, *, variables, dates, title):

        def _name(field, _, original_metadata):
            return self._namer(field, original_metadata)

        data = FieldArray([f.clone(name=_name) for f in data])

        valid_datetime = [_.isoformat() for _ in dates]
        LOG.info("Selecting fields %s %s", len(data), valid_datetime)

        data = data.sel(name=variables, valid_datetime=valid_datetime).order_by(
            name=variables, valid_datetime="ascending"
        )

        check_data(title, data, variables, dates)

        return data

    def _find_variable(self, data, name, **kwargs):
        def _name(field, _, original_metadata):
            return self._namer(field, original_metadata)

        data = FieldArray([f.clone(name=_name) for f in data])
        return data.sel(name=name, **kwargs)

    def _load_forcings(self, fields, variables, dates):
        data = self._filter_and_sort(fields, variables=variables, dates=dates, title="Load forcings")
        return data.to_numpy(dtype=np.float32, flatten=True).reshape(len(variables), len(dates), -1)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

import numpy as np

from ..output import ForwardOutput
from . import create_output
from . import output_registry

LOG = logging.getLogger(__name__)


@output_registry.register("extract_lam")
class ExtractLamOutput(ForwardOutput):
    """_summary_"""

    def __init__(self, context, *, output, lam="lam_0", output_frequency=None, write_initial_state=None):
        super().__init__(context, output_frequency=output_frequency, write_initial_state=write_initial_state)

        if "cutout_mask" in self.checkpoint.supporting_arrays:
            # Backwards compatibility
            mask = self.checkpoint.load_supporting_array("cutout_mask")
            points = slice(None, -np.sum(mask))
        else:
            if lam != "lam_0":
                raise NotImplementedError("Only lam_0 is supported")

            if "lam_1/cutout_mask" in self.checkpoint.supporting_arrays:
                raise NotImplementedError("Only lam_0 is supported")

            mask = self.checkpoint.load_supporting_array(f"{lam}/cutout_mask")
            assert len(mask) == np.sum(mask)
            points = slice(None, np.sum(mask))

        self.points = points
        self.output = create_output(context, output)

    def __repr__(self):
        return f"ExtractLamOutput({self.points}, {self.output})"

    def write_initial_step(self, state):
        # Note: we foreward to 'state', so we write-up options again
        self.output.write_initial_state(self._apply_mask(state))

    def write_step(self, state):
        # Note: we foreward to 'state', so we write-up options again
        self.output.write_state(self._apply_mask(state))

    def _apply_mask(self, state):

        state = state.copy()
        state["fields"] = state["fields"].copy()
        state["latitudes"] = state["latitudes"][self.points]
        state["longitudes"] = state["longitudes"][self.points]

        for field in state["fields"]:
            data = state["fields"][field]
            if data.ndim == 1:
                data = data[self.points]
            else:
                data = data[..., self.points]
            state["fields"][field] = data

        return state

    def close(self):
        self.output.close()

    def print_summary(self, depth=0):
        super().print_summary(depth)
        self.output.print_summary(depth + 1)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import numpy as np

LOG = logging.getLogger(__name__)


class Processor:
    """A processor that transforms a state"""

    def __init__(self, processor):
        self.processor = processor

    def __call__(self, source):
        for state in source:
            yield self.processor(state)


class Chain:
    """A chain of processors"""

    def __init__(self, name, processors):
        self.processors = processors if processors else []
        self.name = name
        LOG.info("%s: %s", self.name, self.processors)

    def __call__(self, source):
        for processor in self.processors:
            source = processor(source)
        return source

    def append(self, processor):
        self.processors.append(processor)
        return self

    def prepend(self, processor):
        self.processors.insert(0, processor)
        return self


class Accumulator:
    """Accumulate fields from zero and return the accumulated fields"""

    def __init__(self, accumulations):
        self.accumulations = accumulations
        LOG.info("Accumulating fields %s", self.accumulations)

        self.accumulators = {}

    def __call__(self, source):
        for state in source:
            for accumulation in self.accumulations:
                if accumulation in state["fields"]:
                    if accumulation not in self.accumulators:
                        self.accumulators[accumulation] = np.zeros_like(state["fields"][accumulation])
                    self.accumulators[accumulation] += np.maximum(0, state["fields"][accumulation])
                    state["fields"][accumulation] = self.accumulators[accumulation]

            yield state



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

from .ekd import EkdInput

LOG = logging.getLogger(__name__)


class GribInput(EkdInput):
    """Handles GRIB input fields."""

    def set_private_attributes(self, state, input_fields):
        # For now we just pass all the fields
        # Later, we can select a relevant subset (e.g. only one
        # level), to save memory

        # By sorting, we will have the most recent field last
        # no we can also use that list to write step 0
        input_fields = input_fields.order_by("valid_datetime")

        state["_grib_templates_for_output"] = {field.metadata("name"): field for field in input_fields}



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from functools import wraps

MARKER = object()


class main_argument:
    """Decorator to set the main argument of a function

    For example...

    @main_argument("path")
    def grib_file_output(context, path, encoding=None, archive_requests=None):
        ...

    So we can have:

    output:
        grib: out.grib

    means the same as

    output:
        grib:
            path: out.grib

    """

    def __init__(self, name):
        self.name = name

    def __call__(self, f):

        @wraps(f)
        def decorator(context, main=MARKER, *args, **kwargs):
            if main is not MARKER:
                kwargs[self.name] = main
            return f(context, *args, **kwargs)

        return decorator



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import numpy as np

LOG = logging.getLogger(__name__)


class Noop:

    def __call__(self, source):
        yield from source


class Accumulator:
    """Accumulate fields from zero and return the accumulated fields"""

    def __init__(self, accumulations):
        self.accumulations = accumulations
        LOG.info("Accumulating fields %s", self.accumulations)

        self.accumulators = {}

    def __call__(self, source):
        for state in source:
            for accumulation in self.accumulations:
                if accumulation in state["fields"]:
                    if accumulation not in self.accumulators:
                        self.accumulators[accumulation] = np.zeros_like(state["fields"][accumulation])
                    self.accumulators[accumulation] += np.maximum(0, state["fields"][accumulation])
                    state["fields"][accumulation] = self.accumulators[accumulation]

            yield state



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import datetime
import logging
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

import yaml
from anemoi.utils.config import merge_configs
from pydantic import BaseModel

LOG = logging.getLogger(__name__)


class Configuration(BaseModel):

    class Config:
        extra = "forbid"

    description: Optional[str] = None

    checkpoint: str | Dict[Literal["huggingface"], Dict[str, Any] | str]
    """A path to an Anemoi checkpoint file."""

    runner: str = "default"
    """The runner to use."""

    date: Union[str, int, datetime.datetime, None] = None
    """The starting date for the forecast. If not provided, the date will depend on the selected Input object. If a string, it is parsed by :func:`anemoi.utils.dates.as_datetime`.
    """

    lead_time: str | int | datetime.timedelta = "10d"
    """The lead time for the forecast. This can be a string, an integer or a timedelta object.
    If an integer, it represents a number of hours. Otherwise, it is parsed by :func:`anemoi.utils.dates.as_timedelta`.
    """

    name: Optional[str] = None
    """Used by prepml."""

    verbosity: int = 0
    """The verbosity level of the runner. This can be 0 (default), 1, 2 or 3."""

    world_size: Optional[int] = 1
    """ Number of parallel processes, used for parallel inference without SLURM """

    report_error: bool = False
    """If True, the runner list the training versions of the packages in case of error."""

    input: Union[str, Dict, None] = "test"
    output: Union[str, Dict, None] = "printer"

    pre_processors: List[Union[str, Dict]] = []
    post_processors: Optional[List[Union[str, Dict]]] = None  # TODO: change default to [] when we have a factory

    forcings: Union[Dict[str, Dict], None] = None
    """Where to find the forcings."""

    device: str = "cuda"
    """The device on which the model should run. This can be "cpu", "cuda" or any other value supported by PyTorch."""

    precision: Optional[str] = None
    """The precision in which the model should run. If not provided, the model will use the precision used during training."""

    allow_nans: Optional[bool] = None
    """
    - If None (default), the model will check for NaNs in the input. If NaNs are found, the model issue a warning and `allow_nans` to True.
    - If False, the model will raise an exception if NaNs are found in the input and output.
    - If True, the model will allow NaNs in the input and output.
    """

    use_grib_paramid: bool = False
    """If True, the runner will use the grib parameter ID when generating MARS requests."""

    write_initial_state: bool = True
    """Wether to write the initial state to the output file. If the model is multi-step, only fields at the forecast reference date are
    written."""

    output_frequency: Optional[str] = None
    """The frequency at which to write the output. This can be a string or an integer. If a string, it is parsed by :func:`anemoi.utils.dates.as_timedelta`."""

    env: Dict[str, str | int] = {}
    """Environment variables to set before running the model. This may be useful to control some packages
    such as `eccodes`. In certain cases, the variables mey be set too late, if the package for which they are intended
    is already loaded when the runner is configured."""

    patch_metadata: dict = {}
    """A dictionary of metadata to patch the checkpoint metadata with. This is used to test new features or to work around
    issues with the checkpoint metadata."""

    development_hacks: dict = {}
    """A dictionary of development hacks to apply to the runner. This is used to test new features or to work around"""

    debugging_info: dict = {}
    """A dictionary to store debug information. This is ignored."""


def load_config(path, overrides, defaults=None, Configuration=Configuration):

    config = {}

    # Set default values
    if defaults is not None:
        if not isinstance(defaults, list):
            defaults = [defaults]
        for d in defaults:
            if isinstance(d, str):
                with open(d) as f:
                    d = yaml.safe_load(f)
            config.update(d)

    # Load the configuration
    with open(path) as f:
        user_config = yaml.safe_load(f)
        config = merge_configs(config, user_config)

    # Apply overrides
    for override in overrides:
        path = config
        key, value = override.split("=")
        keys = key.split(".")
        for key in keys[:-1]:
            path = path.setdefault(key, {})
        path[keys[-1]] = value

    # Validate the configuration
    config = Configuration(**config)

    # Set environment variables found in the configuration
    # as soon as possible
    for key, value in config.env.items():
        os.environ[key] = str(value)

    return config



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import os
import warnings
from collections import defaultdict
from functools import cached_property
from types import MappingProxyType as frozendict
from typing import Literal
from typing import Optional

import numpy as np
from anemoi.transform.variables import Variable
from anemoi.utils.config import DotDict
from anemoi.utils.dates import frequency_to_timedelta as to_timedelta
from anemoi.utils.provenance import gather_provenance_info

from .legacy import LegacyMixin
from .patch import PatchMixin

LOG = logging.getLogger(__name__)


def _remove_full_paths(x):
    if isinstance(x, dict):
        return {k: _remove_full_paths(v) for k, v in x.items()}
    if isinstance(x, list):
        return [_remove_full_paths(v) for v in x]
    if isinstance(x, str):
        return os.path.splitext(os.path.basename(x))[0]
    return x


class Metadata(PatchMixin, LegacyMixin):
    """An object that holds metadata of a checkpoint."""

    def __init__(self, metadata, supporting_arrays={}):
        self._metadata = DotDict(metadata)
        assert isinstance(supporting_arrays, dict)
        self._supporting_arrays = supporting_arrays

    @property
    def _indices(self):
        return self._metadata.data_indices

    @property
    def _config_data(self):
        return self._config.data

    @property
    def _config_training(self):
        return self._config.training

    @property
    def _config_model(self):
        return self._config.model

    @property
    def _config(self):
        return self._metadata.config

    ###########################################################################
    # Debugging
    ###########################################################################

    def _print_indices(self, title, indices, naming, skip=set()):
        LOG.info("")
        LOG.info("%s:", title)

        for k, v in sorted(indices.items()):
            if k in skip:
                continue
            for name, idx in sorted(v.items()):
                entry = f"{k}.{name}"
                if entry in skip:
                    continue

                LOG.info("   %s:", f"{k}.{name}")
                for n in idx:
                    LOG.info(f"     {n:3d} - %s", naming[k].get(n, "?"))
                if not idx:
                    LOG.info("     <empty>")

    def print_indices(self):

        v = {i: v for i, v in enumerate(self.variables)}
        r = {v: k for k, v in self.variable_to_input_tensor_index.items()}
        s = self.output_tensor_index_to_variable

        self._print_indices("Data indices", self._indices.data, dict(input=v, output=v), skip=["output"])
        self._print_indices("Model indices", self._indices.model, dict(input=r, output=s, skip=["output.full"]))

    ###########################################################################
    # Inference
    ###########################################################################

    @cached_property
    def timestep(self):
        """Model time stepping timestep"""
        # frequency = to_timedelta(self._config_data.frequency)
        timestep = to_timedelta(self._config_data.timestep)
        return timestep

    @cached_property
    def precision(self):
        """Return the precision of the model (bits per float)"""
        return self._config_training.precision

    def _make_indices_mapping(self, indices_from, indices_to):
        assert len(indices_from) == len(indices_to)
        return frozendict({i: j for i, j in zip(indices_from, indices_to)})

    @property
    def variable_to_input_tensor_index(self):
        """Return the mapping between variable name and input tensor index"""
        mapping = self._make_indices_mapping(
            self._indices.data.input.full,
            self._indices.model.input.full,
        )

        return frozendict({v: mapping[i] for i, v in enumerate(self.variables) if i in mapping})

    @cached_property
    def output_tensor_index_to_variable(self):
        """Return the mapping between output tensor index and variable name"""
        mapping = self._make_indices_mapping(
            self._indices.model.output.full,
            self._indices.data.output.full,
        )
        return frozendict({k: self.variables[v] for k, v in mapping.items()})

    @cached_property
    def number_of_grid_points(self):
        """Return the number of grid points per fields"""
        if "grid_indices" in self._supporting_arrays:
            return len(self.load_supporting_array("grid_indices"))
        try:
            return self._metadata.dataset.shape[-1]
        except AttributeError:
            return self._legacy_number_of_grid_points()

    @cached_property
    def number_of_input_features(self):
        """Return the number of input features"""
        return len(self._indices.model.input.full)

    @cached_property
    def model_computed_variables(self):
        """The initial conditions variables that need to be computed and not retrieved"""
        typed_variables = self.typed_variables
        return tuple(name for name, v in typed_variables.items() if v.is_computed_forcing)

    @cached_property
    def multi_step_input(self):
        """Number of past steps needed for the initial conditions tensor"""
        return self._config_training.multistep_input

    @cached_property
    def prognostic_output_mask(self):
        return np.array(self._indices.model.output.prognostic)

    @cached_property
    def prognostic_input_mask(self):
        return np.array(self._indices.model.input.prognostic)

    @cached_property
    def computed_time_dependent_forcings(self):
        """Return the indices and names of the computed forcings that are not constant in time"""

        # Mapping between model and data indices
        mapping = self._make_indices_mapping(
            self._indices.model.input.full,
            self._indices.data.input.full,
        )

        # Mapping between model indices and variable names
        forcings_variables = {self.variables[mapping[i]]: i for i in self._indices.model.input.forcing}
        typed_variables = self.typed_variables

        # Filter out the computed forcings that are not constant in time
        indices = []
        variables = []
        for name, idx in sorted(forcings_variables.items(), key=lambda x: x[1]):
            v = typed_variables[name]
            if v.is_computed_forcing and not v.is_constant_in_time:
                indices.append(idx)
                variables.append(name)

        return np.array(indices), variables

    @cached_property
    def computed_constant_forcings(self):
        """Return the indices and names of the computed forcings that are  constant in time"""

        # Mapping between model and data indices
        mapping = self._make_indices_mapping(
            self._indices.model.input.full,
            self._indices.data.input.full,
        )

        # Mapping between model indices and variable names
        forcings_variables = {self.variables[mapping[i]]: i for i in self._indices.model.input.forcing}
        typed_variables = self.typed_variables

        # Filter out the computed forcings that are not constant in time
        indices = []
        variables = []
        for name, idx in sorted(forcings_variables.items(), key=lambda x: x[1]):
            v = typed_variables[name]
            if v.is_computed_forcing and v.is_constant_in_time:
                indices.append(idx)
                variables.append(name)

        return np.array(indices), variables

    ###########################################################################
    # Variables
    ###########################################################################

    @property
    def variables(self):
        """Return the variables as found in the training dataset"""
        return tuple(self._metadata.dataset.variables)

    @cached_property
    def variables_metadata(self):
        """Return the variables and their metadata as found in the training dataset"""
        try:
            result = self._metadata.dataset.variables_metadata
            self._legacy_check_variables_metadata(result)
        except AttributeError:
            result = self._legacy_variables_metadata()

        if "constant_fields" in self._metadata.dataset:
            for name in self._metadata.dataset.constant_fields:
                result[name]["constant_in_time"] = True

        return result

    @cached_property
    def diagnostic_variables(self):
        """Variables that are marked as diagnostic"""
        return [self.index_to_variable[i] for i in self._indices.data.input.diagnostic]

    @cached_property
    def prognostic_variables(self):
        """Variables that are marked as prognostic"""
        return [self.index_to_variable[i] for i in self._indices.data.input.prognostic]

    @cached_property
    def index_to_variable(self):
        """Return a mapping from index to variable name"""
        return frozendict({i: v for i, v in enumerate(self.variables)})

    @cached_property
    def typed_variables(self):
        """Returns a strongly typed variables"""
        result = {name: Variable.from_dict(name, self.variables_metadata[name]) for name in self.variables}

        if "cos_latitude" in result:
            assert result["cos_latitude"].is_computed_forcing
            assert result["cos_latitude"].is_constant_in_time

        if "cos_julian_day" in result:
            assert result["cos_julian_day"].is_computed_forcing
            assert not result["cos_julian_day"].is_constant_in_time

        return result

    @cached_property
    def accumulations(self):
        """Return the indices of the variables that are accumulations"""
        return [v.name for v in self.typed_variables.values() if v.is_accumulation]

    def name_fields(self, fields, namer=None):
        from earthkit.data.indexing.fieldlist import FieldArray

        if namer is None:
            namer = self.default_namer()

        def _name(field, _, original_metadata):
            return namer(field, original_metadata)

        return FieldArray([f.copy(name=_name) for f in fields])

    def sort_by_name(self, fields, namer=None, *args, **kwargs):
        fields = self.name_fields(fields, namer=namer)
        return fields.order_by("name", *args, **kwargs)

    ###########################################################################
    # Default namer
    ###########################################################################

    def default_namer(self, *args, **kwargs):
        """Return a callable that can be used to name earthkit-data fields.
        In that case, return the namer that was used to create the
        training dataset.
        """

        assert len(args) == 0, args
        assert len(kwargs) == 0, kwargs

        def namer(field, metadata):
            warnings.warn("🚧  TEMPORARY CODE 🚧: Use the remapping in the metadata")
            param, levelist, levtype = (
                metadata.get("param"),
                metadata.get("levelist"),
                metadata.get("levtype"),
            )

            # Bug in eccodes that returns levelist for single level fields in GRIB2
            if levtype in ("sfc", "o2d"):
                levelist = None

            if levelist is None:
                return param

            return f"{param}_{levelist}"

        return namer

    ###########################################################################
    # Data retrieval
    ###########################################################################

    @property
    def _data_request(self):
        """Return the data request as encoded in the dataset"""
        try:
            return self._metadata.dataset.data_request
        except AttributeError:
            return self._legacy_data_request()

    @property
    def grid(self):
        return self._data_request.get("grid")

    @property
    def area(self):
        return self._data_request.get("area")

    def variables_from_input(self, *, include_forcings):
        variable_categories = self.variable_categories()
        result = []
        for variable, metadata in self.variables_metadata.items():

            if "mars" not in metadata:
                continue

            if "forcing" in variable_categories[variable]:
                if not include_forcings:
                    continue

            if "computed" in variable_categories[variable]:
                continue

            if "diagnostic" in variable_categories[variable]:
                continue

            result.append(variable)

        return result

    def mars_input_requests(self):
        variable_categories = self.variable_categories()
        for variable in self.variables_from_input(include_forcings=True):

            if "diagnostic" in variable_categories[variable]:
                continue

            metadata = self.variables_metadata[variable]

            yield metadata["mars"].copy()

    def mars_by_levtype(self, levtype):
        variable_categories = self.variable_categories()

        params = set()
        levels = set()

        for variable in self.variables_from_input(include_forcings=True):

            if "diagnostic" in variable_categories[variable]:
                continue

            metadata = self.variables_metadata[variable]

            mars = metadata["mars"]
            if mars.get("levtype") != levtype:
                continue

            if "param" in mars:
                params.add(mars["param"])

            if "levelist" in mars:
                levels.add(mars["levelist"])

        return params, levels

    def mars_requests(self, *, variables, use_grib_paramid=False):
        """Return a list of MARS requests for the variables in the dataset"""

        from anemoi.utils.grib import shortname_to_paramid

        if len(variables) == 0:
            raise ValueError("No variables requested")

        for variable in variables:

            if variable not in self.variables_metadata:
                raise ValueError(f"Variable {variable} not found in the metadata")

            if "mars" not in self.variables_metadata[variable]:
                raise ValueError(f"Variable {variable} has no MARS metadata")

            mars = self.variables_metadata[variable]["mars"].copy()

            for k in ("date", "time"):
                mars.pop(k, None)

            if use_grib_paramid and "param" in mars:
                mars["param"] = shortname_to_paramid(mars["param"])

            yield mars

    ###########################################################################
    # Error reporting
    ###########################################################################

    def report_error(self):

        provenance = self._metadata.provenance_training

        def _print(title, provenance):
            LOG.info("")
            LOG.info("%s:", title)
            for package, git in sorted(provenance.get("git_versions", {}).items()):
                if package.startswith("anemoi."):
                    sha1 = git.get("git", {}).get("sha1", "unknown")
                    LOG.info(f"   {package:20}: {sha1}")

            for package, version in sorted(provenance.get("module_versions", {}).items()):
                if package.startswith("anemoi."):
                    LOG.info(f"   {package:20}: {version}")

        _print("Environment used during training", provenance)
        _print("This environment", gather_provenance_info())

        LOG.warning("If you are running from a git repository, the versions reported above may not be accurate.")
        LOG.warning("The versions are only updated after a `pip install -e .`")

    def validate_environment(
        self,
        *,
        all_packages: bool = False,
        on_difference: Literal["warn", "error", "ignore"] = "warn",
        exempt_packages: Optional[list[str]] = None,
    ) -> bool:
        """Validate environment of the checkpoint against the current environment.

        Parameters
        ----------
        all_packages : bool, optional
            Check all packages in environment or just `anemoi`'s, by default False
        on_difference : Literal['warn', 'error', 'ignore'], optional
            What to do on difference, by default "warn"
        exempt_packages : list[str], optional
            List of packages to exempt from the check, by default EXEMPT_PACKAGES

        Returns
        -------
        bool
            True if environment is valid, False otherwise

        Raises
        ------
        RuntimeError
            If found difference and `on_difference` is 'error'
        ValueError
            If `on_difference` is not 'warn' or 'error'
        """
        from anemoi.inference.provenance import validate_environment

        return validate_environment(
            self, all_packages=all_packages, on_difference=on_difference, exempt_packages=exempt_packages
        )

    ###########################################################################

    def _get_datasets_full_paths(self):
        """This is a temporary method to get the full paths of the datasets used in the training.
        we need to review what goes in the dataset metadata
        """

        result = []

        def _find(x):

            if isinstance(x, list):
                for y in x:
                    if isinstance(y, str):
                        result.append(y)
                    else:
                        _find(y)

            if isinstance(x, dict):
                if "dataset" in x and isinstance(x["dataset"], str):
                    result.append(x["dataset"])

                for k, v in x.items():
                    _find(v)

        _find(self._config.dataloader.training.dataset)
        return result

    def open_dataset_args_kwargs(self, *, use_original_paths, from_dataloader=None):

        # Rebuild the full paths
        # Some older checkpoints may not have the full paths

        full_paths = self._get_datasets_full_paths()
        mapping = {}

        for path in full_paths:
            mapping[os.path.basename(path)] = path
            mapping[os.path.splitext(os.path.basename(path))[0]] = path

        def _fix(x):
            if isinstance(x, list):
                return [_fix(a) for a in x]

            if isinstance(x, dict):
                return {k: _fix(v) for k, v in x.items()}

            if isinstance(x, str):
                return mapping.get(x, x)

            return x

        if from_dataloader is not None:
            args, kwargs = [], self._metadata.config.dataloader[from_dataloader]
        else:
            args, kwargs = self._metadata.dataset.arguments.args, self._metadata.dataset.arguments.kwargs

        args, kwargs = _fix([args, kwargs])

        if use_original_paths:
            return args, kwargs

        return _remove_full_paths(args), _remove_full_paths(kwargs)

    ###########################################################################
    # Not sure this belongs here
    # We need factories
    ###########################################################################

    def variable_categories(self):
        result = defaultdict(set)
        typed_variables = self.typed_variables

        variables_in_data_space = self.variables
        variables_in_model_space = self.output_tensor_index_to_variable

        for name in self._config.data.forcing:
            result[name].add("forcing")

        for idx in self._indices.data.input.diagnostic:
            name = variables_in_data_space[idx]
            result[name].add("diagnostic")

        # assert self._indices.model.input.prognostic == self._indices.model.output.prognostic
        for idx in self._indices.model.output.prognostic:
            name = variables_in_model_space[idx]
            result[name].add("prognostic")

        for name, v in typed_variables.items():
            if v.is_accumulation:
                result[name].add("accumulation")

            if v.is_constant_in_time:
                result[name].add("constant")

            if v.is_computed_forcing:
                result[name].add("computed")

        for name in self.variables:
            if name not in result:
                raise ValueError(f"Variable {name} has no category")

            result[name] = sorted(result[name])

        return result

    def constant_forcings_inputs(self, context, input_state):
        # TODO: this does not belong here

        result = []

        provided_variables = set(input_state["fields"].keys())

        # This will manage the dynamic forcings that are computed
        forcing_mask, forcing_variables = self.computed_constant_forcings

        # Ingore provided variables
        new_forcing_mask = []
        new_forcing_variables = []

        for i, name in zip(forcing_mask, forcing_variables):
            if name not in provided_variables:
                new_forcing_mask.append(i)
                new_forcing_variables.append(name)

        LOG.info(
            "Computed constant forcings: before %s, after %s",
            forcing_variables,
            new_forcing_variables,
        )

        forcing_mask = np.array(new_forcing_mask)
        forcing_variables = new_forcing_variables

        if len(forcing_mask) > 0:
            result.append(
                context.create_constant_computed_forcings(
                    forcing_variables,
                    forcing_mask,
                )
            )

        remaining = (
            set(self._config.data.forcing)
            - set(self.model_computed_variables)
            - set(forcing_variables)
            - provided_variables
        )
        if not remaining:
            return result

        LOG.info("Remaining forcings: %s", remaining)

        # We need the mask of the remaining variable in the model.input space

        mapping = self._make_indices_mapping(
            self._indices.data.input.full,
            self._indices.model.input.full,
        )

        remaining = sorted((mapping[self.variables.index(name)], name) for name in remaining)

        LOG.info("Will get the following from MARS for now: %s", remaining)

        remaining_mask = [i for i, _ in remaining]
        remaining = [name for _, name in remaining]

        forcing = context.create_constant_coupled_forcings(remaining, remaining_mask)

        if forcing is not None:
            # SimpleRunner does not support dynamic forcings
            result.append(forcing)

        return result

    def dynamic_forcings_inputs(self, context, input_state):

        result = []

        # This will manage the dynamic forcings that are computed
        forcing_mask, forcing_variables = self.computed_time_dependent_forcings
        if len(forcing_mask) > 0:
            result.append(
                context.create_dynamic_computed_forcings(
                    forcing_variables,
                    forcing_mask,
                )
            )

        remaining = (
            set(self._config.data.forcing)
            - set(self.model_computed_variables)
            - set([name for name, v in self.typed_variables.items() if v.is_constant_in_time])
        )
        if not remaining:
            return result

        LOG.info("Remaining forcings: %s", remaining)

        # We need the mask of the remaining variable in the model.input space

        mapping = self._make_indices_mapping(
            self._indices.data.input.full,
            self._indices.model.input.full,
        )

        remaining = sorted((mapping[self.variables.index(name)], name) for name in remaining)

        LOG.info("Will get the following from `forcings.dynamic`: %s", remaining)

        remaining_mask = [i for i, _ in remaining]
        remaining = [name for _, name in remaining]

        forcing = context.create_dynamic_coupled_forcings(remaining, remaining_mask)

        if forcing is not None:
            # SimpleRunner does not support dynamic forcings
            result.append(forcing)

        return result

    def boundary_forcings_inputs(self, context, input_state):

        result = []

        if "output_mask" in self._supporting_arrays:
            result.append(
                context.create_boundary_forcings(
                    self.prognostic_variables,
                    self.prognostic_input_mask,
                )
            )
        return result

    ###########################################################################
    # Supporting arrays
    ###########################################################################

    def load_supporting_array(self, name):
        if name not in self._supporting_arrays:
            LOG.error("No supporting array named `%s` found.", name)
            LOG.error("Supporting arrays found:")
            for names in self._supporting_arrays.keys():
                LOG.error("  %s", names)
            raise ValueError(f"Supporting array `{name}` not found")
        return self._supporting_arrays[name]

    @property
    def supporting_arrays(self):
        return self._supporting_arrays

    @property
    def latitudes(self):
        return self._supporting_arrays.get("latitudes")

    @property
    def longitudes(self):
        return self._supporting_arrays.get("longitudes")

    @property
    def grid_points_mask(self):
        # TODO
        return None

    def provenance_training(self):
        """Environmental Configuration when trained"""
        return dict(self._metadata.get("provenance_training", {}))

    def sources(self, path):

        if "sources" not in self._metadata.dataset:
            return []

        import zipfile

        from anemoi.utils.checkpoints import load_supporting_arrays

        ###########################################################################
        # With older metadata, the zarr path are not stored in the metadata
        # we need to fix that
        ###########################################################################

        full_paths = self._get_datasets_full_paths()
        print(full_paths)
        n = 0

        def _fix(x):
            nonlocal n

            if isinstance(x, list):
                [_fix(a) for a in x]

            if isinstance(x, dict):
                if x.get("action", "").startswith("zarr"):
                    print(n, x)
                    path = full_paths[n]
                    n += 1
                    x["path"] = path

                {k: _fix(v) for k, v in x.items()}

        _fix(self._metadata.dataset.sources)

        if n != len(full_paths):
            raise ValueError("Not all paths were fixed")

        ###########################################################################

        sources = []

        with zipfile.ZipFile(path, "r") as zipf:
            for i, source in enumerate(self._metadata.dataset.get("sources", [])):
                entries = {
                    name: self._metadata.supporting_arrays_paths[name] for name in source.get("supporting_arrays", [])
                }
                arrays = load_supporting_arrays(zipf, entries)

                name = source.get("name")
                if name is None:
                    name = f"source{i}"

                sources.append(
                    SourceMetadata(
                        self,
                        name,
                        dict(dataset=source),
                        supporting_arrays=arrays,
                    )
                )

        return sources

    def print_variable_categories(self):
        length = max(len(name) for name in self.variables)
        for name, categories in sorted(self.variable_categories().items()):
            LOG.info(f"   {name:{length}} => {', '.join(categories)}")

    ###########################################################################

    def patch(self, patch):
        """Patch the metadata with the given patch"""

        def merge(main, patch):

            for k, v in patch.items():
                if isinstance(v, dict):
                    if k not in main:
                        main[k] = {}
                    merge(main[k], v)
                else:
                    main[k] = v

        merge(self._metadata, patch)


class SourceMetadata(Metadata):
    """An object that holds metadata of a source. It is only the `dataset` and `supporting_arrays` parts of the metadata.
    The rest is forwarded to the parent metadata object.
    """

    def __init__(self, parent, name, metadata, supporting_arrays={}):
        super().__init__(metadata, supporting_arrays)
        self.parent = parent
        self.name = name

    @property
    def latitudes(self):
        return self._supporting_arrays.get(f"{self.name}/latitudes")

    @property
    def longitudes(self):
        return self._supporting_arrays.get(f"{self.name}/longitudes")

    @property
    def grid_points_mask(self):
        for k, v in self._supporting_arrays.items():
            # TODO: This is a bit of a hack
            if k.startswith(f"{self.name}/") and "mask" in k:
                return v
        return None

    ###########################################################################
    # Forward to parent metadata
    ###########################################################################

    @property
    def _config_training(self):
        return self.parent._config_training

    @property
    def _config_data(self):
        return self.parent._config_data

    @property
    def _indices(self):
        return self.parent._indices

    @property
    def _config(self):
        return self.parent._config

    ###########################################################################



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

from ..output import ForwardOutput
from . import create_output
from . import output_registry

LOG = logging.getLogger(__name__)


@output_registry.register("tee")
class TeeOutput(ForwardOutput):
    """_summary_"""

    def __init__(self, context, *args, outputs=None, output_frequency=None, write_initial_state=None, **kwargs):
        super().__init__(context, output_frequency=output_frequency, write_initial_state=write_initial_state)

        if outputs is None:
            outputs = args

        assert isinstance(outputs, (list, tuple)), outputs
        self.outputs = [create_output(context, output) for output in outputs]

    # We override write_initial_state and write_state
    # so users can configures each levels independently
    def write_initial_state(self, state):
        for output in self.outputs:
            output.write_initial_state(state)

    def write_state(self, state):
        for output in self.outputs:
            output.write_state(state)

    def write_step(self, state):
        raise NotImplementedError("TeeOutput does not support write_step")

    def open(self, state):
        for output in self.outputs:
            output.open(state)

    def close(self):
        for output in self.outputs:
            output.close()

    def __repr__(self):
        return f"TeeOutput({self.outputs})"

    def print_summary(self, depth=0):
        super().print_summary(depth)
        for output in self.outputs:
            output.print_summary(depth + 1)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os

import numpy as np

from ..decorators import main_argument
from ..output import Output
from . import output_registry

LOG = logging.getLogger(__name__)


@output_registry.register("raw")
@main_argument("path")
class RawOutput(Output):
    """_summary_"""

    def __init__(
        self,
        context,
        path,
        template="{date}.npz",
        strftime="%Y%m%d%H%M%S",
        output_frequency=None,
        write_initial_state=None,
    ):
        super().__init__(context, output_frequency=output_frequency, write_initial_state=write_initial_state)
        self.path = path
        self.template = template
        self.strftime = strftime

    def __repr__(self):
        return f"RawOutput({self.path})"

    def write_step(self, state):
        os.makedirs(self.path, exist_ok=True)
        date = state["date"].strftime(self.strftime)
        fn_state = f"{self.path}/{self.template.format(date=date)}"
        restate = {f"field_{key}": val for key, val in state["fields"].items()}
        for key in ["date"]:
            restate[key] = np.array(state[key], dtype=str)
        for key in ["latitudes", "longitudes"]:
            restate[key] = np.array(state[key])
        np.savez_compressed(fn_state, **restate)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os

import numpy as np

from ..decorators import main_argument
from ..output import Output
from . import output_registry

LOG = logging.getLogger(__name__)


@output_registry.register("netcdf")
@main_argument("path")
class NetCDFOutput(Output):
    """_summary_"""

    def __init__(self, context, path, output_frequency=None, write_initial_state=None):
        super().__init__(context, output_frequency=output_frequency, write_initial_state=write_initial_state)
        self.path = path
        self.ncfile = None
        self.float_size = "f4"

    def __repr__(self):
        return f"NetCDFOutput({self.path})"

    def __del__(self):
        if self.ncfile is not None:
            self.ncfile.close()

    def open(self, state):
        from netCDF4 import Dataset

        # If the file exists, we may get a 'Permission denied' error
        if os.path.exists(self.path):
            os.remove(self.path)

        if os.path.dirname(self.path):
            os.makedirs(os.path.dirname(self.path), exist_ok=True)

        self.ncfile = Dataset(self.path, "w", format="NETCDF4")

        compression = {}  # dict(zlib=False, complevel=0)

        values = len(state["latitudes"])

        self.values_dim = self.ncfile.createDimension("values", values)
        self.time_dim = self.ncfile.createDimension("time", size=None)  # infinite dimension
        self.time_var = self.ncfile.createVariable("time", "i8", ("time",), **compression)

        self.time_var.units = "seconds since {0}".format(self.reference_date)
        self.time_var.long_name = "time"
        self.time_var.calendar = "gregorian"

        latitudes = state["latitudes"]
        self.latitude_var = self.ncfile.createVariable(
            "latitude",
            self.float_size,
            ("values",),
            **compression,
        )
        self.latitude_var.units = "degrees_north"
        self.latitude_var.long_name = "latitude"

        longitudes = state["longitudes"]
        self.longitude_var = self.ncfile.createVariable(
            "longitude",
            self.float_size,
            ("values",),
            **compression,
        )
        self.longitude_var.units = "degrees_east"
        self.longitude_var.long_name = "longitude"

        self.vars = {}
        self.ensure_variables(state)

        self.latitude_var[:] = latitudes
        self.longitude_var[:] = longitudes

        self.n = 0
        return self.ncfile

    def ensure_variables(self, state):
        values = len(state["latitudes"])

        compression = {}  # dict(zlib=False, complevel=0)

        for name in state["fields"].keys():
            if name in self.vars:
                continue
            chunksizes = (1, values)

            while np.prod(chunksizes) > 1000000:
                chunksizes = tuple(int(np.ceil(x / 2)) for x in chunksizes)

            self.vars[name] = self.ncfile.createVariable(
                name,
                self.float_size,
                ("time", "values"),
                chunksizes=chunksizes,
                **compression,
            )

    def write_step(self, state):

        self.ensure_variables(state)

        self.time_var[self.n] = self.step(state).total_seconds()

        for name, value in state["fields"].items():
            self.vars[name][self.n] = value

        self.n += 1

    def close(self):
        if self.ncfile is not None:
            self.ncfile.close()
            self.ncfile = None



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json
import logging
from collections import defaultdict

import earthkit.data as ekd
import numpy as np

from ..decorators import main_argument
from ..grib.encoding import check_encoding
from . import output_registry
from .grib import GribOutput

LOG = logging.getLogger(__name__)

# There is a bug with hindcasts, where these keys are not added to the 'mars' namespace
MARS_MAYBE_MISSING_KEYS = (
    "number",
    "step",
    "time",
    "date",
    "hdate",
    "type",
    "stream",
    "expver",
    "class",
    "levtype",
    "levelist",
    "param",
)


def _is_valid(mars, keys):
    if "number" in keys and "number" not in mars:
        LOG.warning("`number` is missing from mars namespace")
        return False

    if "referenceDate" in keys and "hdate" not in mars:
        LOG.warning("`hdate` is missing from mars namespace")
        return False

    return True


class ArchiveCollector:
    """Collects archive requests"""

    UNIQUE = {"date", "hdate", "time", "referenceDate", "type", "stream", "expver"}

    def __init__(self) -> None:
        self.expect = 0
        self._request = defaultdict(set)

    def add(self, field):
        self.expect += 1
        for k, v in field.items():
            self._request[k].add(str(v))
            if k in self.UNIQUE:
                if len(self._request[k]) > 1:
                    raise ValueError(f"Field {field} has different values for {k}: {self._request[k]}")

    @property
    def request(self):
        return {k: sorted(v) for k, v in self._request.items()}


@output_registry.register("grib")
@main_argument("path")
class GribFileOutput(GribOutput):
    """Handles grib files"""

    def __init__(
        self,
        context,
        *,
        path,
        encoding=None,
        archive_requests=None,
        check_encoding=True,
        templates=None,
        grib1_keys=None,
        grib2_keys=None,
        modifiers=None,
        output_frequency=None,
        write_initial_state=None,
        variables=None,
        **kwargs,
    ):
        super().__init__(
            context,
            encoding=encoding,
            templates=templates,
            grib1_keys=grib1_keys,
            grib2_keys=grib2_keys,
            modifiers=modifiers,
            output_frequency=output_frequency,
            write_initial_state=write_initial_state,
            variables=variables,
        )
        self.path = path
        self.output = ekd.new_grib_output(self.path, split_output=True, **kwargs)
        self.archiving = defaultdict(ArchiveCollector)
        self.archive_requests = archive_requests
        self.check_encoding = check_encoding
        self._namespace_bug_fix = False

    def __repr__(self):
        return f"GribFileOutput({self.path})"

    def write_message(self, message, template, **keys):
        # Make sure `name` is not in the keys, otherwise grib_encoding will fail
        if template is not None and template.metadata("name", default=None) is not None:
            # We cannot clear the metadata...
            class Dummy:
                def __init__(self, template):
                    self.template = template
                    self.handle = template.handle

                def __repr__(self):
                    return f"Dummy({self.template})"

            template = Dummy(template)

        # LOG.info("Writing message to %s %s", template, keys)
        try:
            self.collect_archive_requests(
                self.output.write(
                    message,
                    template=template,
                    check_nans=self.context.allow_nans,
                    **keys,
                ),
                template,
                **keys,
            )
        except Exception as e:
            import eccodes

            LOG.error("Error writing message to %s", self.path)
            LOG.error("eccodes: %s", eccodes.__version__)
            LOG.error("Template: %s, Keys: %s", template, keys)
            LOG.error("Exception: %s", e)
            if message is not None and np.isnan(message.data).any():
                LOG.error("Message contains NaNs (%s, %s) (allow_nans=%s)", keys, template, self.context.allow_nans)
            raise

    def collect_archive_requests(self, written, template, **keys):

        if not self.archive_requests and not self.check_encoding:
            return

        handle, path = written

        while True:

            if self._namespace_bug_fix:
                import eccodes
                from earthkit.data.readers.grib.codes import GribCodesHandle

                handle = GribCodesHandle(eccodes.codes_clone(handle._handle), None, None)

            mars = {k: v for k, v in handle.items("mars")}

            if _is_valid(mars, keys):
                break

            if self._namespace_bug_fix:
                raise ValueError("Namespace bug: %s" % mars)

            # Try again with the namespace bug
            LOG.warning("Namespace bug detected, trying again")
            self._namespace_bug_fix = True

        if self.check_encoding:
            check_encoding(handle, keys)

        if self.archive_requests:
            self.archiving[path].add(mars)

    def close(self):
        self.output.close()

        if not self.archive_requests:
            return

        path = self.archive_requests["path"]
        extra = self.archive_requests.get("extra", {})
        patch = self.archive_requests.get("patch", {})
        indent = self.archive_requests.get("indent", None)

        def _patch(r):
            if self.context.config.use_grib_paramid:
                from anemoi.utils.grib import shortname_to_paramid

                param = r.get("param", [])
                if not isinstance(param, list):
                    param = [param]
                param = [shortname_to_paramid(p) for p in param]
                r["param"] = param

            for k, v in patch.items():
                if v is None:
                    r.pop(k, None)
                else:
                    r[k] = v

            return r

        with open(self.archive_requests["path"], "w") as f:
            requests = []

            for path, archive in self.archiving.items():
                assert path is not None, "Path is None"
                request = dict(expect=archive.expect)
                request["source"] = path
                request.update(_patch(archive.request))
                request.update(extra)
                requests.append(request)

            json.dump(requests, f, indent=indent)
            f.write("\n")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json
import logging
from abc import abstractmethod

from earthkit.data.utils.dates import to_datetime

from ..grib.encoding import grib_keys
from ..inputs import create_input
from ..output import Output

LOG = logging.getLogger(__name__)


class HindcastOutput:

    def __init__(self, reference_year):
        self.reference_year = reference_year

    def __call__(self, values, template, keys):

        if "date" not in keys:
            assert template.metadata("hdate", default=None) is None, template
            date = template.metadata("date")
        else:
            date = keys.pop("date")

        for k in ("date", "hdate"):
            keys.pop(k, None)

        keys["edition"] = 1
        keys["localDefinitionNumber"] = 30
        keys["dataDate"] = int(to_datetime(date).strftime("%Y%m%d"))
        keys["referenceDate"] = int(to_datetime(date).replace(year=self.reference_year).strftime("%Y%m%d"))

        return values, template, keys


MODIFIERS = dict(hindcast=HindcastOutput)


def modifier_factory(modifiers):

    if modifiers is None:
        return []

    if not isinstance(modifiers, list):
        modifiers = [modifiers]

    result = []
    for modifier in modifiers:
        assert isinstance(modifier, dict), modifier
        assert len(modifier) == 1, modifier

        klass = list(modifier.keys())[0]
        result.append(MODIFIERS[klass](**modifier[klass]))

    return result


class GribOutput(Output):
    """Handles grib"""

    def __init__(
        self,
        context,
        *,
        encoding=None,
        templates=None,
        grib1_keys=None,
        grib2_keys=None,
        modifiers=None,
        output_frequency=None,
        write_initial_state=None,
        variables=None,
    ):
        super().__init__(context, output_frequency=output_frequency, write_initial_state=write_initial_state)
        self._first = True
        self.typed_variables = self.checkpoint.typed_variables
        self.quiet = set()
        self.encoding = encoding if encoding is not None else {}
        self.templates = templates
        self.grib1_keys = grib1_keys if grib1_keys is not None else {}
        self.grib2_keys = grib2_keys if grib2_keys is not None else {}
        self._template_cache = None
        self._template_source = None
        self._template_date = None
        self._template_reuse = None
        self.use_closest_template = False  # Off for now
        self.modifiers = modifier_factory(modifiers)
        self.variables = variables

    def write_initial_step(self, state):
        # We trust the GribInput class to provide the templates
        # matching the input state

        out_vars = self.variables if self.variables is not None else state["fields"].keys()

        for name in out_vars:

            template = self.template(state, name)
            if template is None:
                # We can currently only write grib output if we have a grib input
                raise ValueError(
                    "GRIB output only works if the input is GRIB (for now). Set `write_initial_step` to `false`."
                )

            variable = self.typed_variables[name]
            if variable.is_accumulation:
                LOG.warning("Found accumulated variable `%s` is initial state.", name)

            # assert False,
            # values = state["fields"][name]
            # assert False, values.shape
            values = None
            keys = grib_keys(
                values=values,
                template=template,
                accumulation=variable.is_accumulation,
                param=None,
                date=None,
                time=None,
                step=0,
                keys=self.encoding,
                grib1_keys=self.grib1_keys,
                grib2_keys=self.grib2_keys,
                quiet=self.quiet,
            )

            for modifier in self.modifiers:
                values, template, keys = modifier(values, template, keys)
            self.write_message(values, template=template, **keys)

    def write_step(self, state):

        reference_date = self.context.reference_date
        date = state["date"]

        step = date - reference_date
        step = step.total_seconds() / 3600
        assert int(step) == step, step
        step = int(step)

        if "_grib_templates_for_output" not in state:
            if "_grib_templates_for_output" not in self.quiet:
                self.quiet.add("_grib_templates_for_output")
                LOG.warning("Input is not GRIB.")

        out_vars = self.variables if self.variables is not None else state["fields"].keys()
        for name in out_vars:
            values = state["fields"][name]
            keys = {}

            variable = self.typed_variables[name]
            param = variable.grib_keys.get("param", variable)

            template = self.template(state, name)

            if template is None:
                if name not in self.quiet:
                    LOG.warning("No GRIB template found for `%s`. This may lead to unexpected results.", name)
                    self.quiet.add(name)

                variable_keys = variable.grib_keys.copy()
                forbidden_keys = ("class", "type", "stream", "expver", "date", "time", "step", "domain")
                for key in forbidden_keys:
                    variable_keys.pop(key, None)

                keys.update(variable_keys)

            keys.update(self.encoding)

            keys = grib_keys(
                values=values,
                template=template,
                date=reference_date.strftime("%Y-%m-%d"),
                time=reference_date.hour,
                step=step,
                param=param,
                accumulation=variable.is_accumulation,
                keys=keys,
                grib1_keys=self.grib1_keys,
                grib2_keys=self.grib2_keys,
                quiet=self.quiet,
            )

            for modifier in self.modifiers:
                values, template, keys = modifier(values, template, keys)

            if LOG.isEnabledFor(logging.DEBUG):
                LOG.info("Encoding GRIB %s\n%s", template, json.dumps(keys, indent=4))

            try:
                self.write_message(values, template=template, **keys)
            except Exception:
                LOG.error("Error writing field %s", name)
                LOG.error("Template: %s", template)
                LOG.error("Keys:\n%s", json.dumps(keys, indent=4))
                raise

    @abstractmethod
    def write_message(self, message, *args, **kwargs):
        pass

    def template(self, state, name):

        if self._template_cache is None:
            self._template_cache = {}
            if "_grib_templates_for_output" in state:
                self._template_cache.update(state.get("_grib_templates_for_output", {}))

        if name is None:
            return self._template_cache

        if name in self._template_cache:
            return self._template_cache[name]

        # Catch all template
        if None in self._template_cache:
            return self._template_cache[None]

        if self.use_closest_template:  #
            template, name2 = self._clostest_template(self._template_cache, name)

            if name not in self.quiet:
                if name2 is not None:
                    LOG.warning("Using template for `%s`", name2)
                self.quiet.add(name)

                self._template_cache[name] = template
                return template

        if not self.templates:
            return None

        if self._template_source is None:
            if "source" not in self.templates:
                raise ValueError("No `source` given in `templates`")

            self._template_source = create_input(self.context, self.templates["source"])
            LOG.info("Loading templates from %s", self._template_source)

            if "date" in self.templates:
                self._template_date = to_datetime(self.templates["date"])

            self._template_reuse = self.templates.get("reuse", False)

        LOG.info("Loading template for %s from %s", name, self._template_source)

        date = self._template_date if self._template_date is not None else state["date"]
        param = self.typed_variables[name].grib_keys.get("param", name)
        field = self._template_source.template(variable=param, date=date)

        if field is None:
            LOG.warning("No template found for `%s`", name)

        if self._template_reuse:
            self._template_cache[None] = field
        else:
            self._template_cache[name] = field
        return field

    def _clostest_template(self, templates, name):
        best = None, None
        best_similarity = 0
        md1 = self.typed_variables[name]
        for name2, template in templates.items():
            md2 = self.typed_variables[name2]
            similarity = md1.similarity(md2)
            if similarity > best_similarity:
                best = template, name2
                best_similarity = similarity
        return best



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json
import sys

from earthkit.data.utils.dates import to_datetime

from ..config import load_config
from ..inputs.grib import GribInput
from ..inputs.mars import postproc
from ..runners import create_runner
from . import Command


def checkpoint_to_requests(
    checkpoint,
    *,
    date,
    target=None,
    include_forcings=True,
    retrieve_fields_type=None,
    staging_dates=None,
    use_grib_paramid=False,
    extra=None,
    use_scda=False,
):
    # TODO: Move this to the runner

    variables = checkpoint.variables_from_input(include_forcings=include_forcings)
    area = checkpoint.area
    grid = checkpoint.grid

    if retrieve_fields_type is not None:
        selected = set()

        for name, kinds in checkpoint.variable_categories().items():
            if "computed" in kinds:
                continue
            for kind in kinds:
                if retrieve_fields_type.startswith(kind):  # PrepML adds an 's' to the type
                    selected.add(name)

        variables = sorted(selected)

    more = postproc(grid, area)

    for r in extra or []:
        k, v = r.split("=")
        more[k] = v

    if staging_dates:
        dates = set()
        with open(staging_dates) as f:
            for line in f:
                date = to_datetime(line.strip())
                dates.update([date + h for h in checkpoint.lagged])
        dates = sorted(dates)
    else:
        date = to_datetime(date)
        dates = [date + h for h in checkpoint.lagged]

    requests = []
    for r in checkpoint.mars_requests(
        dates=dates,
        variables=variables,
        use_grib_paramid=use_grib_paramid,
        always_split_time=use_scda,
    ):
        r = r.copy()
        r.update(more)
        if use_scda:
            _patch_scda(date, r)
        if target is not None:
            r["target"] = target
        requests.append(r)

    return requests


class RetrieveCmd(Command):
    """Used by prepml."""

    need_logging = False

    def add_arguments(self, command_parser):
        command_parser.description = self.__doc__
        command_parser.add_argument("config", type=str, help="Path to config file")
        command_parser.add_argument("--defaults", action="append", help="Sources of default values.")
        command_parser.add_argument("--date", type=str, help="Date")
        command_parser.add_argument("--output", type=str, default=None, help="Output file")
        command_parser.add_argument("--staging-dates", type=str, help="Path to a file with staging dates")
        command_parser.add_argument("--extra", action="append", help="Additional request values. Can be repeated")
        command_parser.add_argument("--retrieve-fields-type", type=str, help="Type of fields to retrieve")
        command_parser.add_argument("--use-scda", action="store_true", help="Use scda stream for 6/18 input time")
        command_parser.add_argument("overrides", nargs="*", help="Overrides.")

    def run(self, args):

        config = load_config(args.config, args.overrides, defaults=args.defaults)

        runner = create_runner(config)

        # so that the user does not need to pass --extra target=path when the input file is already in the config
        target = None
        input = runner.create_input()
        if isinstance(input, GribInput) and (path := getattr(input, "path", None)):
            target = path

        requests = checkpoint_to_requests(
            runner.checkpoint,
            date=args.date,
            target=target,
            include_forcings=True,
            extra=args.extra,
            retrieve_fields_type=args.retrieve_fields_type,
            staging_dates=args.staging_dates,
            use_grib_paramid=config.use_grib_paramid,
            use_scda=args.use_scda,
        )

        if args.output and args.output != "-":
            f = open(args.output, "w")
        else:
            f = sys.stdout

        json.dump(requests, f, indent=4)


def _patch_scda(base_date, request):
    if base_date.hour not in (6, 18):
        return

    if request.get("class", "od") != "od":
        return
    if request.get("type", "an") not in ("an", "fc"):
        return
    if request.get("stream", "oper") not in ("oper", "scda"):
        return

    request_time = int(request.get("time", 12))
    if request_time > 100:
        request_time = int(request_time / 100)

    if request_time in (6, 18):
        request["stream"] = "scda"


command = RetrieveCmd



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

from ..output import ForwardOutput
from . import create_output
from . import output_registry

LOG = logging.getLogger(__name__)


@output_registry.register("apply_mask")
class ApplyMaskOutput(ForwardOutput):
    """_summary_"""

    def __init__(self, context, *, mask, output, output_frequency=None, write_initial_state=None):
        super().__init__(context, output_frequency=output_frequency, write_initial_state=write_initial_state)
        self.mask = self.checkpoint.load_supporting_array(mask)
        self.output = create_output(context, output)

    def __repr__(self):
        return f"ApplyMaskOutput({self.mask}, {self.output})"

    def write_initial_step(self, state):
        # Note: we foreward to 'state', so we write-up options again
        self.output.write_initial_state(self._apply_mask(state))

    def write_step(self, state):
        # Note: we foreward to 'state', so we write-up options again
        self.output.write_state(self._apply_mask(state))

    def _apply_mask(self, state):
        state = state.copy()
        state["fields"] = state["fields"].copy()
        state["latitudes"] = state["latitudes"][self.mask]
        state["longitudes"] = state["longitudes"][self.mask]

        for field in state["fields"]:
            data = state["fields"][field]
            if data.ndim == 1:
                data = data[self.mask]
            else:
                data = data[..., self.mask]
            state["fields"][field] = data

        return state

    def close(self):
        self.output.close()

    def print_summary(self, depth=0):
        super().print_summary(depth)
        self.output.print_summary(depth + 1)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import earthkit.data as ekd
import numpy as np

from . import input_registry
from .grib import GribInput

LOG = logging.getLogger(__name__)


@input_registry.register("icon_grib_file")
class IconInput(GribInput):
    """Handles grib files from ICON
    WARNING: this code will become a pugin in the future
    """

    def __init__(self, context, path, grid, refinement_level_c, namer=None, **kwargs):
        super().__init__(context, namer=namer, **kwargs)
        self.path = path
        self.grid = grid
        self.refinement_level_c = refinement_level_c

    def create_input_state(self, *, date):
        import xarray as xr

        LOG.info(f"Reading ICON grid from {self.grid}")
        ds = xr.open_dataset(self.grid)
        latitudes = np.rad2deg(ds.clat[ds.refinement_level_c <= self.refinement_level_c].values)
        longitudes = np.rad2deg(ds.clon[ds.refinement_level_c <= self.refinement_level_c].values)

        LOG.info(f"Latitudes {np.min(latitudes)} {np.max(latitudes)}")
        LOG.info(f"Longitudes {np.min(longitudes)} {np.max(longitudes)}")

        LOG.info("Done")

        return self._create_input_state(
            ekd.from_source("file", self.path),
            variables=None,
            date=date,
            latitudes=latitudes,
            longitudes=longitudes,
        )

    def load_forcings(self, *, variables, dates):
        return self._load_forcings(ekd.from_source("file", self.path), variables=variables, dates=dates)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

from ..forcings import ComputedForcings
from ..forcings import Forcings
from ..runner import Runner
from . import runner_registry

LOG = logging.getLogger(__name__)

# This is because forcings are assumed to already be in the
# state dictionary, so we don't need to load them from anywhere.


class NoForcings(Forcings):
    """No forcings."""

    def __init__(self, context, variables, mask):
        super().__init__(context)
        self.variables = variables
        self.mask = mask
        self.kinds = dict(unknown=True)

    def load_forcings(self, state, date):
        pass


@runner_registry.register("simple")
class SimpleRunner(Runner):
    """Use that runner when using the low level API."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def create_constant_computed_forcings(self, variables, mask):
        result = ComputedForcings(self, variables, mask)
        LOG.info("Constant computed forcing: %s", result)
        return result

    def create_dynamic_computed_forcings(self, variables, mask):
        result = ComputedForcings(self, variables, mask)
        LOG.info("Dynamic computed forcing: %s", result)
        return result

    def create_constant_coupled_forcings(self, variables, mask):
        # This runner does not support coupled forcings
        # there are supposed to be already in the state dictionary
        # of managed by the user.
        LOG.warning("Coupled forcings are not supported by this runner: %s", variables)
        return None

    def create_dynamic_coupled_forcings(self, variables, mask):
        # This runner does not support coupled forcings
        # there are supposed to be already in the state dictionary
        # of managed by the user.
        LOG.warning("Coupled forcings are not supported by this runner: %s", variables)
        return None



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import warnings

from anemoi.utils.config import DotDict
from pydantic import BaseModel

from ..forcings import BoundaryForcings
from ..forcings import ComputedForcings
from ..forcings import CoupledForcings
from ..inputs import create_input
from ..outputs import create_output
from ..runner import Runner
from . import runner_registry

LOG = logging.getLogger(__name__)


@runner_registry.register("default")
class DefaultRunner(Runner):
    """Runner from a configuration file."""

    def __init__(self, config):

        if isinstance(config, dict):
            # So we get the dot notation
            config = DotDict(config)

        # Remove that when the Pydantic model is ready
        if isinstance(config, BaseModel):
            config = DotDict(config.model_dump())

        self.config = config

        # temporary code
        accumulate_from_start_of_forecast = _check_accumulation_processor(config)

        super().__init__(
            config.checkpoint,
            device=config.device,
            precision=config.precision,
            allow_nans=config.allow_nans,
            verbosity=config.verbosity,
            report_error=config.report_error,
            use_grib_paramid=config.use_grib_paramid,
            patch_metadata=config.patch_metadata,
            development_hacks=config.development_hacks,
            output_frequency=config.output_frequency,
            write_initial_state=config.write_initial_state,
            accumulate_from_start_of_forecast=accumulate_from_start_of_forecast,
        )

    def create_input(self):
        input = create_input(self, self.config.input)
        LOG.info("Input: %s", input)
        return input

    def create_output(self):
        output = create_output(self, self.config.output)
        LOG.info("Output:")
        output.print_summary()
        return output

    # Computed forcings
    def create_constant_computed_forcings(self, variables, mask):
        result = ComputedForcings(self, variables, mask)
        LOG.info("Constant computed forcing: %s", result)
        return result

    def create_dynamic_computed_forcings(self, variables, mask):
        result = ComputedForcings(self, variables, mask)
        LOG.info("Dynamic computed forcing: %s", result)
        return result

    # Coupled forcings
    # TODO: Connect them to the Input if needed
    # Here, by default, we may use the same input "class" as the input
    # not the same instance. This means that we may call mars several times

    def _input_forcings(self, name):
        if self.config.forcings is None:
            # Use the same as the input
            return self.config.input

        if name in self.config.forcings:
            return self.config.forcings[name]

        if "input" in self.config.forcings:
            return self.config.forcings.input

        return self.config.forcings

    def create_constant_coupled_forcings(self, variables, mask):
        input = create_input(self, self._input_forcings("constant"))
        result = CoupledForcings(self, input, variables, mask)
        LOG.info("Constant coupled forcing: %s", result)
        return result

    def create_dynamic_coupled_forcings(self, variables, mask):

        input = create_input(self, self._input_forcings("dynamic"))
        result = CoupledForcings(self, input, variables, mask)
        LOG.info("Dynamic coupled forcing: %s", result)
        return result

    def create_boundary_forcings(self, variables, mask):

        input = create_input(self, self._input_forcings("boundary"))
        result = BoundaryForcings(self, input, variables, mask)
        LOG.info("Boundary forcing: %s", result)
        return result


def _check_accumulation_processor(config):
    # TODO #131: Remove this when we have a processor factory
    # For now, implement a three-way switch.
    # post_processors: None -> accumulate_from_start_of_forecast = True
    # post_processors: []   -> accumulate_from_start_of_forecast = False
    # post_processors: ["accumulate_from_start_of_forecast"] -> accumulate_from_start_of_forecast = True
    pre_processors = config.get("pre_processors")
    post_processors = config.get("post_processors")

    if pre_processors:
        raise NotImplementedError("pre_processors are not yet supported. Please remove this entry from your config.")

    if post_processors not in (None, [], ["accumulate_from_start_of_forecast"]):
        raise ValueError("post_processors only supports `accumulate_from_start_of_forecast`.")

    if isinstance(post_processors, list):
        accumulate_from_start_of_forecast = "accumulate_from_start_of_forecast" in post_processors

        if not accumulate_from_start_of_forecast:
            warnings.warn(
                """
                post_processors are defined but `accumulate_from_start_of_forecast` is not set."
                🚧 Accumulations will NOT be accumulated from the beginning of the forecast. 🚧
                """
            )
    else:
        warnings.warn(
            """
            No post_processors defined. Accumulations will be accumulated from the beginning of the forecast.

            🚧🚧🚧 In a future release, the default will be to NOT accumulate from the beginning of the forecast. 🚧🚧🚧
            Update your config if you wish to keep accumulating from the beginning.
            https://github.com/ecmwf/anemoi-inference/issues/131
            """,
        )
        accumulate_from_start_of_forecast = True

    LOG.info("accumulate_from_start_of_forecast: %s", accumulate_from_start_of_forecast)
    return accumulate_from_start_of_forecast



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import TYPE_CHECKING
from typing import Literal
from typing import Optional

from anemoi.utils.provenance import gather_provenance_info
from packaging.version import Version

if TYPE_CHECKING:
    from anemoi.inference.metadata import Metadata


# Complete package name to be exempt
EXEMPT_PACKAGES = [
    "anemoi.training",
    "hydra",
    "hydra_plugins",
    "lightning",
    "pytorch_lightning",
    "lightning_fabric",
    "lightning_utilities",
]

# Entire namespaces to be exempt
EXEMPT_NAMESPACES = [
    "hydra_plugins",
]

LOG = logging.getLogger(__name__)


def validate_environment(
    metadata: "Metadata",
    *,
    all_packages: bool = False,
    on_difference: Literal["warn", "error", "ignore"] = "warn",
    exempt_packages: Optional[list[str]] = None,
) -> bool:
    """Validate environment of the checkpoint against the current environment.

    Parameters
    ----------
    metadata : Metadata
        Metadata object of the checkpoint, to validate against
    all_packages : bool, optional
        Check all packages in environment or just `anemoi`'s, by default False
    on_difference : Literal['warn', 'error', 'ignore'], optional
        What to do on difference, by default "warn"
    exempt_packages : list[str], optional
        List of packages to exempt from the check, by default EXEMPT_PACKAGES

    Returns
    -------
    bool
        True if environment is valid, False otherwise

    Raises
    ------
    RuntimeError
        If found difference and `on_difference` is 'error'
    ValueError
        If `on_difference` is not 'warn' or 'error'
    """
    train_environment = metadata.provenance_training()
    inference_environment = gather_provenance_info(full=False)

    # Override module information with more complete inference environment capture
    import importlib.metadata as imp_metadata

    module_versions = {
        distribution.metadata.get("Name", "").replace("-", "_"): distribution.metadata["Version"]
        for distribution in imp_metadata.distributions()
    }

    inference_environment["module_versions"] = module_versions

    exempt_packages = exempt_packages or []
    exempt_packages.extend(EXEMPT_PACKAGES)

    invalid_messages = {
        "python": [],
        "missing": [],
        "mismatch": [],
        "critical mismatch": [],
        "uncommitted": [],
    }

    if train_environment["python"] != inference_environment["python"]:
        invalid_messages["python"].append(
            f"Python version mismatch: {train_environment['python']} != {inference_environment['python']}"
        )

    for module in train_environment["module_versions"].keys():
        inference_module_name = module  # Due to package name differences between retrieval methods this may change

        if not all_packages and "anemoi" not in module:
            continue
        elif module in exempt_packages or module.split(".")[0] in EXEMPT_NAMESPACES:
            continue
        elif module.startswith("_"):
            continue
        elif module not in inference_environment["module_versions"]:
            if "." in module and module.replace(".", "_") in inference_environment["module_versions"]:
                inference_module_name = module.replace(".", "_")
            else:
                try:
                    import importlib

                    importlib.import_module(module)
                    continue
                except (ModuleNotFoundError, ImportError):
                    pass
                invalid_messages["missing"].append(f"Missing module in inference environment: {module}")
                continue

        train_environment_version = Version(train_environment["module_versions"][module])
        inference_environment_version = Version(inference_environment["module_versions"][inference_module_name])

        if train_environment_version < inference_environment_version:
            invalid_messages["mismatch"].append(
                f"Version of module {module} was lower in training then in inference: {train_environment_version!s} <= {inference_environment_version!s}"
            )
        elif train_environment_version > inference_environment_version:
            invalid_messages["critical mismatch"].append(
                f"CRITICAL: Version of module {module} was greater in training then in inference: {train_environment_version!s} > {inference_environment_version!s}"
            )

    for git_record in train_environment["git_versions"].keys():
        file_record = train_environment["git_versions"][git_record]["git"]
        if file_record["modified_files"] == 0 and file_record["untracked_files"] == 0:
            continue

        if git_record not in inference_environment["git_versions"]:
            invalid_messages["uncommitted"].append(
                f"Training environment contained uncommitted change missing in inference environment: {git_record}"
            )
        elif (
            train_environment["git_versions"][git_record]["sha1"]
            != inference_environment["git_versions"][git_record]["sha1"]
        ):
            invalid_messages["uncommitted"].append(
                f"sha1 mismatch for git record between training and inference. {git_record} (training != inference): {train_environment['git_versions'][git_record]} != {inference_environment['git_versions'][git_record]}"
            )

    for git_record in inference_environment["git_versions"].keys():
        file_record = inference_environment["git_versions"][git_record]["git"]
        if file_record["modified_files"] == 0 and file_record["untracked_files"] == 0:
            continue

        if git_record not in train_environment["git_versions"]:
            invalid_messages["uncommitted"].append(
                f"Inference environment contains uncommited changes missing in training: {git_record}"
            )

    if len(invalid_messages) > 0:
        text = "Environment validation failed. The following issues were found:\n" + "\n".join(
            [f"  {key}:\n    " + "\n    ".join(value) for key, value in invalid_messages.items() if len(value) > 0]
        )
        if on_difference == "warn":
            LOG.warning(text)
        elif on_difference == "error":
            raise RuntimeError(text)
        elif on_difference == "ignore":
            pass
        else:
            raise ValueError(f"Invalid value for `on_difference`: {on_difference}")
        return False

    LOG.info("Environment validation passed")
    return True



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

from ..context import Context
from ..input import Input
from ..runner import Runner
from . import runner_registry

LOG = logging.getLogger(__name__)


class CutoutInput(Input):
    """An Input object that combines two inputs."""

    def __init__(self, context, lam, globe):
        super().__init__(context)
        self.lam = lam
        self.globe = globe

    def create_input_state(self, *, date=None):

        state1 = self.lam.create_input_state(date=date)
        state2 = self.globe.create_input_state(date=date)

        assert False, (state1, state2)


class CutoutContext(Context):
    """A Context object for CutoutRunner."""

    def __init__(self, checkpoint):
        self._checkpoint = checkpoint

    @property
    def checkpoint(self):
        return self._checkpoint


@runner_registry.register("cutout")
class CutoutRunner(Runner):
    """A Runner that for LAMs."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        sources = self.checkpoint.sources

        if len(sources) != 2:
            raise ValueError(f"CutoutRunner expects two source, found {len(sources)}.")

        self.lam, self.globe = sources



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

from earthkit.data.utils.dates import to_datetime

from . import input_registry
from .grib import GribInput
from .mars import postproc

LOG = logging.getLogger(__name__)


def retrieve(requests, grid, area, dataset, **kwargs):
    import earthkit.data as ekd

    def _(r):
        mars = r.copy()
        for k, v in r.items():
            if isinstance(v, (list, tuple)):
                mars[k] = "/".join(str(x) for x in v)
            else:
                mars[k] = str(v)

        return ",".join(f"{k}={v}" for k, v in mars.items())

    pproc = postproc(grid, area)

    result = ekd.from_source("empty")
    for r in requests:
        if isinstance(dataset, str):
            d = dataset
        elif isinstance(dataset, dict):
            # Get dataset from intersection of keys between request and dataset dict
            search_dataset = dataset.copy()
            while isinstance(search_dataset, dict):
                keys = set(r.keys()).intersection(set(search_dataset.keys()))
                if len(keys) == 0:
                    raise KeyError(
                        f"While searching for dataset, could not find any valid key in dictionary: {r.keys()}, {search_dataset}"
                    )
                key = list(keys)[0]
                if r[key] not in search_dataset[key]:
                    if "*" in search_dataset[key]:
                        search_dataset = search_dataset[key]["*"]
                        continue

                    raise KeyError(
                        f"Dataset dictionary does not contain key {r[key]!r} in {key!r}: {dict(search_dataset[key])}."
                    )
                search_dataset = search_dataset[key][r[key]]

            d = search_dataset

        r.update(pproc)
        r.update(kwargs)

        LOG.debug("%s", _(r))
        result += ekd.from_source("cds", d, r)

    return result


@input_registry.register("cds")
class CDSInput(GribInput):
    """Get input fields from CDS"""

    def __init__(self, context, *, dataset, namer=None, **kwargs):
        super().__init__(context, namer=namer)

        self.variables = self.checkpoint.variables_from_input(include_forcings=False)
        self.dataset = dataset
        self.kwargs = kwargs

    def create_input_state(self, *, date):
        if date is None:
            date = to_datetime(-1)
            LOG.warning("CDSInput: `date` parameter not provided, using yesterday's date: %s", date)

        date = to_datetime(date)

        return self._create_input_state(
            self.retrieve(
                self.variables,
                [date + h for h in self.checkpoint.lagged],
            ),
            variables=self.variables,
            date=date,
        )

    def retrieve(self, variables, dates):

        requests = self.checkpoint.mars_requests(
            variables=variables,
            dates=dates,
            use_grib_paramid=self.context.use_grib_paramid,
        )

        if not requests:
            raise ValueError("No requests for %s (%s)" % (variables, dates))

        return retrieve(
            requests, self.checkpoint.grid, self.checkpoint.area, dataset=self.dataset, expver="0001", **self.kwargs
        )

    def template(self, variable, date, **kwargs):
        return self.retrieve([variable], [date])[0]

    def load_forcings(self, variables, dates):
        return self._load_forcings(self.retrieve(variables, dates), variables, dates)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging
from abc import ABC
from abc import abstractmethod

LOG = logging.getLogger(__name__)


class Context(ABC):
    """Represents the context of the inference."""

    allow_nans = (None,)  # can be True of False
    use_grib_paramid = False
    verbosity = 0
    development_hacks = {}  # For testing purposes, don't use in production

    # Some runners will set these values, which can be queried by Output objects,
    # but may remain as None

    reference_date = None
    time_step = None
    lead_time = None
    output_frequency = None
    write_initial_state = True

    ##################################################################

    @property
    @abstractmethod
    def checkpoint(self):
        """Returns the checkpoint used for the inference."""
        pass

    ##################################################################
    # The methods below are not marked as abstract because they are
    # not required to be implemented by subclasses. However, they are
    # expected to be implemented by subclasses when relevant.
    # For example, when running the `SimpleRunner`, the user is
    # expected to provide the forcings directly as input to the runner.
    ##################################################################

    def create_input(self):
        raise NotImplementedError()

    def create_output(self):
        raise NotImplementedError()

    def create_constant_computed_forcings(self, variables, mask):
        raise NotImplementedError(f"{self.__class__.__name__}.create_constant_computed_forcings")

    def create_constant_coupled_forcings(self, variables, mask):
        raise NotImplementedError(f"{self.__class__.__name__}.create_constant_coupled_forcings")

    def create_dynamic_computed_forcings(self, variables, mask):
        raise NotImplementedError(f"{self.__class__.__name__}.create_dynamic_computed_forcings")

    def create_dynamic_coupled_forcings(self, variables, mask):
        raise NotImplementedError(f"{self.__class__.__name__}.create_dynamic_coupled_forcings")
